/*
	Copyright (C) 2004 - 2005 rt2x00 SourceForge Project
	<http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
	Module Name: rt2x00core.c

	Abstract: rt2x00 core routines.
	Supported chipsets: RT2460, RT2560 & RT2570.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		3 Sep 2005	Initial Release (Beta 1).
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>

#include "rt2x00.h"
#include "rt2x00core.h"

#ifdef DRV_NAME
#undef DRV_NAME
#define DRV_NAME			"rt2x00core"
#endif /* DRV_NAME */

struct _rt2x00_device*
rt2x00_device(struct net_device *net_dev)
{
	return ieee80211_priv(net_dev);
}
EXPORT_SYMBOL_GPL(rt2x00_device);

static void
rt2x00_update_config(struct _rt2x00_device *device, u16 update_flags)
{
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(test_and_set_bit(DEVICE_CONFIG_UPDATE, &device->flags))
		return;

	down_write(&device->rt2x00_sem);

	update_flags |= core->config.update_flags;
	core->config.update_flags = 0;

	downgrade_write(&device->rt2x00_sem);

	if(core->config.update_flags);
		device->handler->dev_update_config(device, &core->config, update_flags);

	up_read(&device->rt2x00_sem);

	clear_bit(DEVICE_CONFIG_UPDATE, &device->flags);
}


/*
 * Management frames handlers
 */
#define MGMT_FRAME_SIZE			256

static inline struct sk_buff*
rt2x00_allocate_mgmt_frame(unsigned int length)
{
	struct sk_buff				*skb = NULL;

	skb = dev_alloc_skb(MGMT_FRAME_SIZE);
	if(!skb){
		ERROR("Not enough memory to create mgmt frame.\n");
		return NULL;
	}

	skb_reserve(skb, NET_IP_ALIGN);

	if(unlikely(!skb_put(skb, length))){
		dev_kfree_skb_any(skb);
		ERROR("Padding of mgmt frame failed.\n");
		return NULL;
	}

	return skb;
}

static int
rt2x00_encrypt_mgmt_frame(struct _rt2x00_device *device, struct sk_buff *skb)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct ieee80211_crypt_data		*crypt = NULL;
	struct sk_buff				*mgmt = NULL;
	int					result = 0x00000000;

	spin_lock(&core->ieee80211->lock);

	crypt = core->ieee80211->crypt[core->ieee80211->tx_keyidx];

	mgmt = skb_copy_expand(skb,
		crypt->ops->extra_msdu_prefix_len + crypt->ops->extra_mpdu_prefix_len,
		crypt->ops->extra_msdu_postfix_len + crypt->ops->extra_mpdu_postfix_len,
		GFP_KERNEL);
	if(!mgmt){
		INFO("Failed to expand buffer for encryption.\n");
		spin_unlock(&core->ieee80211->lock);
		return -ENOMEM;
	}

	dev_kfree_skb_any(skb);
	skb = mgmt;

	atomic_inc(&crypt->refcnt);

	if(crypt->ops->encrypt_msdu)
		result = crypt->ops->encrypt_msdu(skb, IEEE80211_3ADDR_LEN, crypt->priv);

	atomic_dec(&crypt->refcnt);

	if(result < 0){
		INFO("Encryption (msdu) of mgmt frame failed.\n");
		core->ieee80211->ieee_stats.tx_discards++;
		spin_unlock(&core->ieee80211->lock);
		return -EFAULT;
	}

	if (core->ieee80211->tkip_countermeasures
	&& crypt && crypt->ops && strcmp(crypt->ops->name, "TKIP") == 0){
		if(net_ratelimit())
			INFO("TKIP countermeasures, mgmt frame dropped.\n");
		spin_unlock(&core->ieee80211->lock);
		return 0;
	}

	atomic_inc(&crypt->refcnt);

	if(crypt->ops->encrypt_mpdu)
		result = crypt->ops->encrypt_mpdu(skb, IEEE80211_3ADDR_LEN, crypt->priv);

	atomic_dec(&crypt->refcnt);

	if(result < 0){
		INFO("Encryption (mpdu) of mgmt frame failed.\n");
		core->ieee80211->ieee_stats.tx_discards++;
		spin_unlock(&core->ieee80211->lock);
		return -EFAULT;
	}

	skb_put(skb, 4);

	spin_unlock(&core->ieee80211->lock);

	return 0;
}

static void
rt2x00_start_xmit_mgmt(struct _rt2x00_device *device, struct _packet_ring *ring, struct sk_buff *skb)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct ieee80211_hdr_3addr		*ieee80211hdr = (struct ieee80211_hdr_3addr*)skb->data;

	if(ring->type != RING_BEACON)
		ieee80211hdr->seq_ctl = cpu_to_le16(rt2x00_get_sequence(device));

	spin_lock(&ring->lock);

	skb_queue_tail(&ring->queue, skb);

	if(ring->ready)
		queue_work(core->workqueue, &ring->work);

	spin_unlock(&ring->lock);
}

/*
 * The wpa_cipher array is required for the MFIE_TYPE_GENERIC element.
 * wpa_cipher[9], wpa_cipher[15] and wpa_cipher[21] should be filled in
 * later since those values depend on the encryption type.
 */
#if WIRELESS_EXT > 17
static const u8 wpa_cipher[22] = {
		0x00, 0x50, 0xf2, 0x01,			/* Oui. */
		0x01, 0x00,				/* Version. */
		0x00, 0x50, 0xf2, 0x00,			/* Multicast. */
		0x01, 0x00,				/* Number of unicast. */
		0x00, 0x50, 0xf2, 0x00,			/* Unicast. */
		0x01, 0x00,				/* Number of authentication method. */
		0x00, 0x50, 0xf2, 0x00			/* Authentication. */
};

static const u8 wpa2_cipher[20] = {
		0x01, 0x00,				/* Version. */
		0x00, 0x0f, 0xac, 0x00,			/* Group cipher. */
		0x01, 0x00,				/* Number of pairwise. */
		0x00, 0x0f, 0xac, 0x00,			/* Unicast. */
		0x01, 0x00,				/* Number of authentication method. */
		0x00, 0x0f, 0xac, 0x00,			/* Authentication. */
		0x00, 0x00,				/* RSN capability. */
};
#endif /* WIRELESS_EXT > 17 */

static int
rt2x00_create_mgmt_buffer(struct _rt2x00_device *device, struct sk_buff *skb)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct ieee80211_info_element		element;
	u8					*data = NULL;
	u8					index = 0x00;

	element.id = MFIE_TYPE_SSID;
	element.len = strlen(&core->config.active.essid[0]);
	data = skb_put(skb, sizeof(struct ieee80211_info_element) + element.len);
	if(!data)
		return -ENOMEM;
	memcpy(data, &element, sizeof(struct ieee80211_info_element));
	data += sizeof(struct ieee80211_info_element);
	memcpy(data, &core->config.active.essid, element.len);

	element.id = MFIE_TYPE_RATES;
	if(core->config.active.network == IEEE_B){
		element.len = 4;
		index = 0;
	}else if(core->config.active.network & IEEE_B){
		element.len = 12;
		index = 0;
	}else{
		element.len = 8;
		index = 4;
	}
	data = skb_put(skb, sizeof(struct ieee80211_info_element) + element.len);
	if(!data)
		return -ENOMEM;
	memcpy(data, &element, sizeof(struct ieee80211_info_element));
	data += sizeof(struct ieee80211_info_element);
	memcpy(data, &capabilities.bitrate[index], element.len);

	element.id = MFIE_TYPE_DS_SET;
	element.len = 1;
	data = skb_put(skb, sizeof(struct ieee80211_info_element) + element.len);
	if(!data)
		return -ENOMEM;
	memcpy(data, &element, sizeof(struct ieee80211_info_element));
	data += sizeof(struct ieee80211_info_element);
	memcpy(data, &core->config.active.channel, element.len);

	if(core->ieee80211->sec.level == SEC_LEVEL_0
	|| core->ieee80211->sec.level == SEC_LEVEL_1)
		return 0;

#if WIRELESS_EXT > 17
	if(core->config.active.wpa_version ==  IW_AUTH_WPA_VERSION_WPA){
		element.id = MFIE_TYPE_GENERIC;
		element.len = ARRAY_SIZE(wpa_cipher);
		data = skb_put(skb, sizeof(struct ieee80211_info_element) + element.len);
		if(!data)
			return -ENOMEM;
		memcpy(data, &element, sizeof(struct ieee80211_info_element));
		data += sizeof(struct ieee80211_info_element);
		memcpy(data, &wpa_cipher, element.len);
		if(core->ieee80211->sec.level == SEC_LEVEL_2){
			*(data + 9) = 0x02;
			*(data + 15) = 0x02;
		}else if(core->ieee80211->sec.level == SEC_LEVEL_3){
			*(data + 9) = 0x04;
			*(data + 15) = 0x04;
		}

		else if(core->config.active.mgmt_auth == IW_AUTH_KEY_MGMT_PSK)
			*(data + 21) = 0x02;
		else
			*(data + 21) = 0x01;
	}else if(core->config.active.wpa_version ==  IW_AUTH_WPA_VERSION_WPA2){
		element.id = MFIE_TYPE_RSN;
		element.len = ARRAY_SIZE(wpa2_cipher);
		data = skb_put(skb, sizeof(struct ieee80211_info_element) + element.len);
		if(!data)
			return -ENOMEM;
		memcpy(data, &element, sizeof(struct ieee80211_info_element));
		data += sizeof(struct ieee80211_info_element);
		memcpy(data, &wpa2_cipher, element.len);
		if(core->ieee80211->sec.level == SEC_LEVEL_2){
			*(data + 5) = 0x02;
			*(data + 11) = 0x02;
		}else if(core->ieee80211->sec.level == SEC_LEVEL_3){
			*(data + 5) = 0x04;
			*(data + 11) = 0x04;
		}

		else if(core->config.active.mgmt_auth == IW_AUTH_KEY_MGMT_PSK)
			*(data + 17) = 0x02;
		else
			*(data + 17) = 0x01;
	}
#endif /* WIRELESS_EXT > 17 */
	return 0;
}

static u16
rt2x00_create_mgmt_capability(struct _rt2x00_device *device)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	u16					capability = 0x0000;

	if(core->config.active.iw_mode == IW_MODE_INFRA)
		capability |= WLAN_CAPABILITY_ESS;
	if(core->config.active.iw_mode == IW_MODE_ADHOC)
		capability |= WLAN_CAPABILITY_IBSS;
	if(core->ieee80211->sec.encrypt)
		capability |= WLAN_CAPABILITY_PRIVACY;
	if(core->config.active.flags & CONFIG_ENABLED_SHORT_PREAMBLE)
		capability |= WLAN_CAPABILITY_SHORT_PREAMBLE;
	if(core->config.slot_time == 9)
		capability |= WLAN_CAPABILITY_SHORT_SLOT_TIME;

	return capability;
}

static void
rt2x00_mgmt_snd_beacon(struct _rt2x00_device *device)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct sk_buff				*skb = NULL;
	struct ieee80211_beacon			*beacon = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_beacon));
	if(!skb){
		ERROR("Failed to send beacon.\n");
		return;
	}

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	beacon = (struct ieee80211_beacon*)skb->data;

	beacon->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_BEACON);
	beacon->header.duration_id = cpu_to_le16(0);
	memset(&beacon->header.addr1, 0xff, ETH_ALEN);
	memcpy(&beacon->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memset(&beacon->header.addr3, 0xff, ETH_ALEN);

	beacon->time_stamp[0] = jiffies;
	beacon->beacon_interval = core->config.beacon;
	beacon->capability = cpu_to_le16(rt2x00_create_mgmt_capability(device));
	if(rt2x00_create_mgmt_buffer(device, skb)){
		ERROR("Failed to send beacon.\n");
		spin_unlock(&core->ieee80211->lock);
		up_read(&device->rt2x00_sem);
		return;
	}

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	spin_lock(&core->beacon.lock);
	skb_queue_purge(&core->beacon.queue);
	spin_unlock(&core->beacon.lock);

	rt2x00_start_xmit_mgmt(device, &core->beacon, skb);
}

static void
rt2x00_mgmt_snd_probe_req(struct _rt2x00_device *device)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct ieee80211_probe_request		*probe = NULL;
	struct sk_buff				*skb = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_probe_request));
	if(!skb){
		ERROR("Failed to send probe request.\n");
		return;
	}

	probe = (struct ieee80211_probe_request*)skb->data;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	probe->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_PROBE_REQ);
	probe->header.duration_id = cpu_to_le16(0);
	memset(&probe->header.addr1, 0xff, ETH_ALEN);
	memcpy(&probe->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memset(&probe->header.addr3, 0xff, ETH_ALEN);

	if(rt2x00_create_mgmt_buffer(device, skb)){
		ERROR("Failed to send probe request.\n");
		spin_unlock(&core->ieee80211->lock);
		up_read(&device->rt2x00_sem);
		return;
	}

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	rt2x00_start_xmit_mgmt(device, &core->prio, skb);

	INFO("Probe request send for channel %d.\n", core->config.active.channel);
}

static void
rt2x00_mgmt_snd_association(struct _rt2x00_device *device, u8 *addr)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct sk_buff				*skb = NULL;
	struct ieee80211_assoc_request		*assoc = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_assoc_request));
	if(!skb){
		ERROR("Failed to send association request.\n");
		return;
	}

	assoc = (struct ieee80211_assoc_request*)skb->data;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	if(core->network && core->network->last_associate)
		assoc->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_REASSOC_REQ);
	else
		assoc->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ASSOC_REQ);
	assoc->header.duration_id = cpu_to_le16(0);
	memcpy(&assoc->header.addr1, addr, ETH_ALEN);
	memcpy(&assoc->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memcpy(&assoc->header.addr3, addr, ETH_ALEN);

	assoc->capability = cpu_to_le16(rt2x00_create_mgmt_capability(device));
	assoc->listen_interval = cpu_to_le16(3);
	//memcpy(&assoc->current_ap, &core->config.active.bssid, ETH_ALEN);

	if(rt2x00_create_mgmt_buffer(device, skb)){
		ERROR("Failed to send association request.\n");
		spin_unlock(&core->ieee80211->lock);
		up_read(&device->rt2x00_sem);
		return;
	}

	memcpy(&core->connect.bssid, addr, ETH_ALEN);
	core->connect.timestamp = jiffies;
	core->connect.status = CONNECT_ASSOCIATE;

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	INFO("Association request send to " ADDR_STRING ".\n", ADDR_PRINT(assoc->header.addr1));

	rt2x00_start_xmit_mgmt(device, &core->prio, skb);
}

static void
rt2x00_mgmt_snd_association_resp(struct _rt2x00_device *device, u8 *addr)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct sk_buff				*skb = NULL;
	struct ieee80211_assoc_response		*assoc = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_assoc_response));
	if(!skb){
		ERROR("Failed to send association response.\n");
		return;
	}

	assoc = (struct ieee80211_assoc_response*)skb->data;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	if(core->network && core->network->last_associate)
		assoc->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_REASSOC_RESP);
	else
		assoc->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ASSOC_RESP);
	assoc->header.duration_id = cpu_to_le16(0);
	memcpy(&assoc->header.addr1, addr, ETH_ALEN);
	memcpy(&assoc->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memcpy(&assoc->header.addr3, addr, ETH_ALEN);

	assoc->capability = cpu_to_le16(rt2x00_create_mgmt_capability(device));
	assoc->status = cpu_to_le16(WLAN_STATUS_SUCCESS);

	if(rt2x00_create_mgmt_buffer(device, skb)){
		ERROR("Failed to send association response.\n");
		spin_unlock(&core->ieee80211->lock);
		up_read(&device->rt2x00_sem);
		return;
	}

	memcpy(&core->connect.bssid, addr, ETH_ALEN);
	core->connect.timestamp = jiffies;
	core->connect.status = CONNECT_ASSOCIATE;

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	INFO("Association response send to " ADDR_STRING ".\n", ADDR_PRINT(assoc->header.addr1));

	rt2x00_start_xmit_mgmt(device, &core->prio, skb);
}

static void
rt2x00_mgmt_rcv_assoc(struct _rt2x00_device *device, struct ieee80211_hdr_4addr *ieee80211hdr, struct ieee80211_rx_stats *rx_stats)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct ieee80211_assoc_request		*assoc = (struct ieee80211_assoc_request*)ieee80211hdr;

	down_read(&device->rt2x00_sem);

	if(core->config.active.flags & CONFIG_FIX_BSSID
	&& memcmp(&assoc->header.addr2, &core->config.active.bssid, ETH_ALEN)){
		up_read(&device->rt2x00_sem);
		return;
	}

	INFO("Association request received from for " ADDR_STRING ".\n", ADDR_PRINT(assoc->header.addr2));
	up_read(&device->rt2x00_sem);

	spin_lock(&core->ieee80211->lock);
	core->ieee80211->state = IEEE80211_ASSOCIATED;
	spin_unlock(&core->ieee80211->lock);

	rt2x00_mgmt_snd_association_resp(device, &assoc->header.addr2[0]);

	return;
}

static void
rt2x00_mgmt_snd_deassociation(struct _rt2x00_device *device, u8 *addr, u16 reason)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct sk_buff				*skb = NULL;
	struct ieee80211_assoc_request		*assoc = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_assoc_request));
	if(!skb){
		ERROR("Failed to send de-association request.\n");
		return;
	}

	assoc = (struct ieee80211_assoc_request*)skb->data;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	assoc->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_DISASSOC);
	assoc->header.duration_id = cpu_to_le16(0);
	memcpy(&assoc->header.addr1, addr, ETH_ALEN);
	memcpy(&assoc->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memcpy(&assoc->header.addr3, addr, ETH_ALEN);

	assoc->capability = cpu_to_le16(reason);

	if(rt2x00_create_mgmt_buffer(device, skb)){
		ERROR("Failed to send de-association request.\n");
		spin_unlock(&core->ieee80211->lock);
		up_read(&device->rt2x00_sem);
		return;
	}

	core->ieee80211->state = IEEE80211_AUTHENTICATED;

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	INFO("De-association request send to " ADDR_STRING ".\n", ADDR_PRINT(assoc->header.addr1));

	rt2x00_start_xmit_mgmt(device, &core->prio, skb);
}

static void
rt2x00_mgmt_snd_authentication(struct _rt2x00_device *device, u8 *addr)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct sk_buff				*skb = NULL;
	struct ieee80211_auth			*auth = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_auth));
	if(!skb){
		ERROR("Failed to send authentication request.\n");
		return;
	}

	auth = (struct ieee80211_auth*)skb->data;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	auth->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_AUTH);
	auth->header.duration_id = cpu_to_le16(0);
	memcpy(&auth->header.addr1, addr, ETH_ALEN);
	memcpy(&auth->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memcpy(&auth->header.addr3, addr, ETH_ALEN);

	auth->algorithm = cpu_to_le16(core->ieee80211->sec.auth_mode);
	auth->transaction = cpu_to_le16(1);
	auth->status = cpu_to_le16(WLAN_STATUS_SUCCESS);

	core->ieee80211->state = IEEE80211_AUTHENTICATING;

	core->connect.timestamp = jiffies;
	core->connect.status = CONNECT_AUTHENTICATE;
	memcpy(&core->connect.bssid, addr, ETH_ALEN);

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	INFO("Authentication request send to " ADDR_STRING ".\n", ADDR_PRINT(auth->header.addr1));

	rt2x00_start_xmit_mgmt(device, &core->prio, skb);
}

static void
rt2x00_mgmt_snd_handshake(struct _rt2x00_device *device, u8 *addr)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct sk_buff				*skb = NULL;
	struct ieee80211_auth			*auth = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_auth));
	if(!skb){
		ERROR("Failed to send authentication handshake.\n");
		return;
	}

	auth = (struct ieee80211_auth*)skb->data;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	auth->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_AUTH);
	auth->header.duration_id = cpu_to_le16(0);
	memcpy(&auth->header.addr1, addr, ETH_ALEN);
	memcpy(&auth->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memcpy(&auth->header.addr3, addr, ETH_ALEN);

	auth->algorithm = cpu_to_le16(core->ieee80211->sec.auth_mode);
	auth->transaction = cpu_to_le16(1);
	auth->status = cpu_to_le16(WLAN_STATUS_SUCCESS);

	core->ieee80211->state = IEEE80211_AUTHENTICATING;

	core->connect.timestamp = jiffies;
	core->connect.status = CONNECT_HANDSHAKE;

	if(rt2x00_encrypt_mgmt_frame(device, skb)){
		ERROR("Failed to send authentication handshake.\n");
		spin_unlock(&core->ieee80211->lock);
		up_read(&device->rt2x00_sem);
	}

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	INFO("Authentication handshake send to " ADDR_STRING ".\n", ADDR_PRINT(auth->header.addr1));

	rt2x00_start_xmit_mgmt(device, &core->prio, skb);
}

static void
rt2x00_mgmt_snd_deauthentication(struct _rt2x00_device *device, u8 *addr, u16 reason)
{
	struct _rt2x00_core			*core = rt2x00_core(device);
	struct sk_buff				*skb = NULL;
	struct ieee80211_auth			*auth = NULL;

	skb = rt2x00_allocate_mgmt_frame(sizeof(struct ieee80211_auth));
	if(!skb){
		ERROR("Failed to send de-authentication request.\n");
		return;
	}

	auth = (struct ieee80211_auth*)skb->data;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	auth->header.frame_ctl = cpu_to_le16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_DEAUTH);
	auth->header.duration_id = cpu_to_le16(0);
	memcpy(&auth->header.addr1, addr, ETH_ALEN);
	memcpy(&auth->header.addr2, &device->net_dev->dev_addr, ETH_ALEN);
	memcpy(&auth->header.addr3, addr, ETH_ALEN);

	auth->algorithm = cpu_to_le16(core->ieee80211->sec.auth_mode);
	auth->transaction = cpu_to_le16(reason);
	auth->status = cpu_to_le16(WLAN_STATUS_SUCCESS);

	core->ieee80211->state = IEEE80211_INITIALIZED;

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	INFO("De-authentication request send to " ADDR_STRING ".\n", ADDR_PRINT(auth->header.addr1));

	rt2x00_start_xmit_mgmt(device, &core->prio, skb);
}

static int
rt2x00_handle_auth(struct net_device *net_dev, struct ieee80211_auth *auth)
{
	struct _rt2x00_device			*device = ieee80211_priv(net_dev);
	struct _rt2x00_core			*core = rt2x00_core(device);
	u16					status = le16_to_cpu(auth->status);
	u16					wpa_handshake = 0x0000;

	spin_lock(&core->ieee80211->lock);
	wpa_handshake = !core->ieee80211->ieee802_1x
			&& core->ieee80211->sec.encrypt
			&& core->ieee80211->sec.level >= SEC_LEVEL_2;
	spin_unlock(&core->ieee80211->lock);

	down_read(&device->rt2x00_sem);

	if(core->config.active.flags & CONFIG_FIX_BSSID
	&& memcmp(&auth->header.addr2, &core->config.active.bssid, ETH_ALEN)){
		up_read(&device->rt2x00_sem);
		return 0;
	}

	if(core->connect.status != CONNECT_AUTHENTICATE
	&& core->connect.status != CONNECT_HANDSHAKE){
		up_read(&device->rt2x00_sem);
		INFO("Authentication request received from " ADDR_STRING ".\n", ADDR_PRINT(auth->header.addr2));
		rt2x00_mgmt_snd_authentication(device, &auth->header.addr2[0]);
		return 0;
	}

	if(status != WLAN_STATUS_SUCCESS){
		INFO("Authentication request failed for " ADDR_STRING ", reason: %d.\n", ADDR_PRINT(auth->header.addr2), status);
		up_read(&device->rt2x00_sem);
		return 0;
	}

	if(core->connect.status == CONNECT_AUTHENTICATE && wpa_handshake){
		core->connect.timestamp = jiffies;
		up_read(&device->rt2x00_sem);
		rt2x00_mgmt_snd_handshake(device, &auth->header.addr2[0]);
		return 0;
	}

	INFO("Authentication request succeeded for " ADDR_STRING ".\n", ADDR_PRINT(auth->header.addr2));
	up_read(&device->rt2x00_sem);

	spin_lock(&core->ieee80211->lock);
	core->ieee80211->state = IEEE80211_AUTHENTICATED;
	spin_unlock(&core->ieee80211->lock);

	rt2x00_mgmt_snd_association(device, &auth->header.addr2[0]);

	return 0;
}

static int
rt2x00_handle_disassoc(struct net_device *net_dev, struct ieee80211_disassoc *assoc)
{
	struct _rt2x00_device			*device = ieee80211_priv(net_dev);
	struct _rt2x00_core			*core = rt2x00_core(device);

	if(core->network
	&& memcmp(&core->network->bssid, &assoc->header.addr2, ETH_ALEN)){
		INFO("Invalid de-association request received from " ADDR_STRING ".\n", ADDR_PRINT(assoc->header.addr2));
		return 0;
	}

	INFO("De-association request received from " ADDR_STRING ".\n", ADDR_PRINT(assoc->header.addr2));

	rt2x00_mgmt_snd_deassociation(device, &assoc->header.addr2[0], assoc->reason_code);

	return 0;
}

static int
rt2x00_handle_assoc_response(struct net_device *net_dev, struct ieee80211_assoc_response *resp, struct ieee80211_network *network)
{
	struct _rt2x00_device			*device = ieee80211_priv(net_dev);
	struct _rt2x00_core			*core = rt2x00_core(device);
	u16					status = le16_to_cpu(resp->status);

	down_read(&device->rt2x00_sem);

	if(core->connect.status != CONNECT_ASSOCIATE
	|| memcmp(&resp->header.addr2, &core->connect.bssid, ETH_ALEN)){
		up_read(&device->rt2x00_sem);
		return 0;
	}

	if(status != WLAN_STATUS_SUCCESS){
		INFO("Association request failed for " ADDR_STRING ", reason: %d.\n", ADDR_PRINT(resp->header.addr2), status);
		up_read(&device->rt2x00_sem);
		return 0;
	}

	INFO("Association request succeeded for " ADDR_STRING ".\n", ADDR_PRINT(resp->header.addr2));

	core->connect.status = CONNECT_SUCCESS;

	network->last_associate = jiffies;
	memcpy(core->network, network, sizeof(*network));

	up_read(&device->rt2x00_sem);

	spin_lock(&core->ieee80211->lock);
	core->ieee80211->state = IEEE80211_ASSOCIATED;
	spin_unlock(&core->ieee80211->lock);

	return 0;
}

/*
 * Scanning routines.
 */
#define SCAN_TIME			( HZ / 5 )
#define BEACON_EXPIRE			( HZ * 2 )

static void
rt2x00_start_scan(struct _rt2x00_device *device, const u16 flags)
{
	struct _rt2x00_core		*core = rt2x00_core(device);
	const struct ieee80211_geo	*geo = ieee80211_get_geo(core->ieee80211);
	u8				channel = 0x00;

	if(test_and_set_bit(DEVICE_SCANNING, &device->flags)){
		NOTICE("Device already busy scanning.\n");
		return;
	}

	core->scan_flags = flags;

	spin_lock(&core->ieee80211->lock);
	if(core->ieee80211->freq_band & IEEE80211_24GHZ_BAND)
		channel = geo->bg[0].channel;
	else if(geo->a_channels > 0)
		channel = geo->a[0].channel;
	spin_unlock(&core->ieee80211->lock);

	if(core->scan_flags & IW_SCAN_ALL_FREQ){
		INFO("Start scanning all channels...\n");
		down_write(&device->rt2x00_sem);
		core->config.active.channel = channel;
		core->config.update_flags |= UPDATE_CHANNEL;
		up_write(&device->rt2x00_sem);
	} else {
		INFO("Start scanning current channel...\n");
	}

	rt2x00_update_config(device, core->config.update_flags);

	rt2x00_mgmt_snd_probe_req(device);

	queue_delayed_work(core->workqueue, &core->scan_tuner, SCAN_TIME);
}

static void
rt2x00_stop_scan(struct _rt2x00_device *device)
{
	struct _rt2x00_core		*core = rt2x00_core(device);

	if(!test_and_clear_bit(DEVICE_SCANNING, &device->flags)){
		NOTICE("Device not busy scanning.\n");
		return;
	}

	if(core->scan_flags & IW_SCAN_ALL_FREQ){
		down_write(&device->rt2x00_sem);
		core->config.active.channel = core->config.user.channel;
		core->config.update_flags |= UPDATE_CHANNEL;
		up_write(&device->rt2x00_sem);
		rt2x00_update_config(device, core->config.update_flags);
	}

	INFO("...Scan ended.\n");
}

static void
rt2x00_scan_periodic(void *data)
{
	struct _rt2x00_device		*device = data;
	struct _rt2x00_core		*core = rt2x00_core(device);
	const struct ieee80211_geo	*geo = ieee80211_get_geo(core->ieee80211);
	u8				index = 0x00;
	u8				channel = 0x00;
	u8				new_channel = 0x00;

	down_read(&device->rt2x00_sem);
	channel = core->config.active.channel;
	up_read(&device->rt2x00_sem);

	if(!test_bit(DEVICE_SCANNING, &device->flags)){
		WARNING("Scan update failed, not busy scanning.\n");
		return;
	}

	spin_lock(&core->ieee80211->lock);

	if(!(core->scan_flags & IW_SCAN_ALL_FREQ)){
		spin_unlock(&core->ieee80211->lock);
		rt2x00_stop_scan(device);
		return;
	}

	index = ieee80211_channel_to_index(core->ieee80211, channel);

	if(core->ieee80211->freq_band & IEEE80211_24GHZ_BAND && channel <= 14){
		if(index < geo->bg_channels - 1)
			new_channel = geo->bg[index + 1].channel;
		else if(core->ieee80211->freq_band & IEEE80211_52GHZ_BAND && geo->a_channels)
			new_channel = geo->a[0].channel;
	}else if(core->ieee80211->freq_band & IEEE80211_52GHZ_BAND && geo->a_channels){
		if(index < geo->a_channels - 1)
			new_channel = geo->a[index + 1].channel;
	}

	spin_unlock(&core->ieee80211->lock);

	if(!new_channel){
		rt2x00_stop_scan(device);
		return;
	}

	down_write(&device->rt2x00_sem);
	core->config.active.channel = new_channel;
	core->config.update_flags |= UPDATE_CHANNEL;
	up_write(&device->rt2x00_sem);

	rt2x00_update_config(device, core->config.update_flags);

	rt2x00_mgmt_snd_probe_req(device);

	queue_delayed_work(core->workqueue, &core->scan_tuner, SCAN_TIME);
}

/*
 * rt2x00_validate_scan_entry should be called with ieee80211 lock held!.
 */
static inline char
rt2x00_validate_scan_entry(struct ieee80211_device *ieee80211, struct _rt2x00_config *config, struct ieee80211_network *entry)
{
	char				status = 0x00;

	status += time_after(jiffies, (entry->last_scanned + BEACON_EXPIRE));

	if(config->active.flags & CONFIG_FIX_BSSID)
		status += memcmp(&config->active.bssid, &entry->bssid, ETH_ALEN);

	if(config->active.flags & CONFIG_FIX_ESSID){
		if(ieee80211_is_empty_essid(&entry->ssid[0], entry->ssid_len))
			status += 0x00;
		else if(entry->ssid_len == strlen(&config->active.essid[0]))
			status += memcmp(&config->active.essid, &entry->ssid, entry->ssid_len);
		else
			status += 1;
	}

	if(config->active.flags & CONFIG_FIX_CHANNEL)
		status += !(entry->channel == config->active.channel);

	if(!(entry->flags & NETWORK_HAS_CCK && ieee80211->modulation & IEEE80211_CCK_MODULATION)
	&& !(entry->flags & NETWORK_HAS_OFDM && ieee80211->modulation & IEEE80211_OFDM_MODULATION))
		status += 1;

	if(((entry->capability & WLAN_CAPABILITY_PRIVACY) ^ ieee80211->sec.encrypt)
	&& (!(entry->capability & WLAN_CAPABILITY_PRIVACY) && !ieee80211->open_wep))
		status += 1;

	return (status == 0x00);
}

/*
 * rt2x00_search_scan should be called with ieee80211 lock held!.
 */
static struct ieee80211_network*
rt2x00_search_scan(struct _rt2x00_device *device, u8 mode)
{
	struct _rt2x00_core		*core = rt2x00_core(device);
	struct ieee80211_network	*result = NULL;
	struct ieee80211_network	*entry = NULL;

	if(test_bit(DEVICE_SCANNING, &device->flags))
		return NULL;

	list_for_each_entry(entry, &core->ieee80211->network_list, list){
		if((mode == IW_MODE_AUTO || entry->mode == mode)
		&& rt2x00_validate_scan_entry(core->ieee80211, &core->config, entry))
			if(!result || entry->stats.rssi > result->stats.rssi)
				result = entry;
	}

	return result;
}

/*
 * Ring control routines.
 */
static inline void
rt2x00_init_ring(struct _rt2x00_device *device, struct _packet_ring *ring, u8 ring_type, void (*handler)(void*))
{
	spin_lock_init(&ring->lock);

	skb_queue_head_init(&ring->queue);

	INIT_WORK(&ring->work, handler, ring);

	ring->core = rt2x00_core(device);
	ring->type = ring_type;
	ring->ready = 0;
}

static inline void
rt2x00_enable_ring(struct _packet_ring *ring)
{
	spin_lock(&ring->lock);
	ring->ready = 1;
	spin_unlock(&ring->lock);
}

static inline void
rt2x00_disable_ring(struct _packet_ring *ring)
{
	spin_lock(&ring->lock);
	ring->ready = 0;
	cancel_delayed_work(&ring->work);
	skb_queue_purge(&ring->queue);
	spin_unlock(&ring->lock);
}

static inline void
rt2x00_kick_ring(struct _packet_ring *ring)
{
	spin_lock(&ring->lock);
	ring->ready = 1;
	queue_work(ring->core->workqueue, &ring->work);
	spin_unlock(&ring->lock);
}

/*
 * Configuration validation routines.
 * The validation functions must be called with the rt2x00_sem held.
 */
static int
rt2x00_validate_config_mode(struct _rt2x00_device *device, struct _rt2x00_config_request *config, u8 iw_mode)
{
	if(iw_mode == IW_MODE_ADHOC && rt2x00_is_ofdm_bitrate(config->bitrate)){
		NOTICE("OFDM rates are now allowed on an adhoc network.\n");
		return -EINVAL;
	}

	return 0;
}

static int
rt2x00_validate_config_channel(struct _rt2x00_device *device, struct _rt2x00_config_request *config, u8 channel)
{
	if((!(config->network & IEEE_A) && rt2x00_is_52ghz_channel(channel))
	|| (config->network == IEEE_A && rt2x00_is_24ghz_channel(channel))){
		NOTICE("Channel not supported for current network type.\n");
		return -EINVAL;
	}

	if(rt2x00_is_52ghz_channel(channel) && rt2x00_is_cck_bitrate(config->bitrate)){
		NOTICE("CCK rates are now allowed on a 5.2GHz channel.\n");
		return -EINVAL;
	}

	if(channel == 14 && rt2x00_is_ofdm_bitrate(config->bitrate)){
		NOTICE("OFDM rates are now allowed on channel 14.\n");
		return -EINVAL;
	}

	return 0;
}

static int
rt2x00_validate_config_bitrate(struct _rt2x00_device *device, struct _rt2x00_config_request *config, u8 bitrate)
{
	struct _rt2x00_core		*core = rt2x00_core(device);
	u8				cck_rate = rt2x00_is_cck_bitrate(bitrate);
	u8				ofdm_rate = rt2x00_is_ofdm_bitrate(bitrate);
	u8				counter = 0x00;

	if(!(cck_rate ^ ofdm_rate)){
		NOTICE("Rate not CCK and not OFDM.\n");
		return -EINVAL;
	}

	if((!(config->network & IEEE_B) && cck_rate)
	|| (config->network == IEEE_B && ofdm_rate)){
		NOTICE("Rate not supported for current network type.\n");
		return -EINVAL;
	}

	if(rt2x00_is_52ghz_channel(config->channel) && cck_rate){
		NOTICE("CCK rates are now allowed on a 5.2GHz channel.\n");
		return -EINVAL;
	}

	if(config->channel == 14 && ofdm_rate){
		NOTICE("OFDM rates are now allowed on channel 14.\n");
		return -EINVAL;
	}


	if(config->iw_mode == IW_MODE_ADHOC && ofdm_rate){
		NOTICE("OFDM rates are now allowed on an adhoc network.\n");
		return -EINVAL;
	}

	if(!core->network)
		return 0;

	for(counter = 0; counter < MAX_RATES_EX_LENGTH; counter++){
		if(counter < core->network->rates_len){
			if(bitrate == (core->network->rates[counter]^IEEE80211_BASIC_RATE_MASK))
				return 0;
		}
		if(counter < core->network->rates_ex_len){
			if(bitrate == (core->network->rates_ex[counter]^IEEE80211_BASIC_RATE_MASK))
				return 0;
		}
	}

	return -EINVAL;
}

static int
rt2x00_activate_config_network(struct _rt2x00_device *device, u8 network)
{
	struct _rt2x00_core		*core = rt2x00_core(device);

	spin_lock(&core->ieee80211->lock);

	core->ieee80211->mode = network;
	core->ieee80211->abg_true = network;

	if(network == IEEE_A){		/* 802.11a */
		core->ieee80211->freq_band &= ~IEEE80211_24GHZ_BAND;
		core->ieee80211->freq_band |= IEEE80211_52GHZ_BAND;
		core->ieee80211->modulation |= IEEE80211_OFDM_MODULATION;
	}else if(network & IEEE_A){	/* 802.11abg */
		core->ieee80211->freq_band |= IEEE80211_24GHZ_BAND;
		core->ieee80211->freq_band |= IEEE80211_52GHZ_BAND;
		core->ieee80211->modulation |= IEEE80211_CCK_MODULATION;
		core->ieee80211->modulation |= IEEE80211_OFDM_MODULATION;
	}else if(network & IEEE_G){	/* 802.11bg */
		core->ieee80211->freq_band |= IEEE80211_24GHZ_BAND;
		core->ieee80211->freq_band &= ~IEEE80211_52GHZ_BAND;
		core->ieee80211->modulation |= IEEE80211_CCK_MODULATION;
		core->ieee80211->modulation |= IEEE80211_OFDM_MODULATION;
	}else{				/* 802.11b */
		core->ieee80211->freq_band |= IEEE80211_24GHZ_BAND;
		core->ieee80211->freq_band &= ~IEEE80211_52GHZ_BAND;
		core->ieee80211->modulation |= IEEE80211_CCK_MODULATION;
		core->ieee80211->modulation &= ~IEEE80211_OFDM_MODULATION;
	}

	core->config.active.network = network;

	spin_unlock(&core->ieee80211->lock);

	return 0;
}

static int
rt2x00_activate_config_geography(struct _rt2x00_device *device, u8 geography)
{
	struct _rt2x00_core		*core = rt2x00_core(device);

	spin_lock(&core->ieee80211->lock);

	ieee80211_set_geo(core->ieee80211, &geography_data[geography]);

	spin_unlock(&core->ieee80211->lock);

	return 0;
}

static void
rt2x00_activate_config(struct _rt2x00_device *device)
{
	struct _rt2x00_core		*core = rt2x00_core(device);

	INFO("Activating configuration...\n");

	down_write(&device->rt2x00_sem);

	if(core->config.user_commit & COMMIT_NETWORK){
		if(!core->config.user.network){
			if(test_bit(DEVICE_CAP_802_11A, &device->flags))
				core->config.user.network |= IEEE_A;
			if(test_bit(DEVICE_CAP_802_11B, &device->flags))
				core->config.user.network |= IEEE_B;
			if(test_bit(DEVICE_CAP_802_11G, &device->flags))
				core->config.user.network |= IEEE_G;
		}
		rt2x00_activate_config_network(device, core->config.user.network);
		core->config.active.network = core->config.user.network;
		core->config.user_commit |= COMMIT_BITRATE | COMMIT_CHANNEL;
	}

	if(core->config.user_commit & COMMIT_GEOGRAPHY){
		if(core->config.user.geography >= IOCTL_GEOGRAPHY_INVAL)
			core->config.user.geography = IOCTL_GEOGRAPHY_INVAL;
		rt2x00_activate_config_geography(device, core->config.user.geography);
		core->config.active.geography = core->config.user.geography;
		core->config.user_commit |= COMMIT_CHANNEL;
	}

	if(core->config.user_commit & COMMIT_MODE){
		if(!rt2x00_validate_config_mode(device, &core->config.active, core->config.user.iw_mode))
			core->config.active.iw_mode = core->config.user.iw_mode;
		else
			core->config.active.iw_mode = IW_MODE_AUTO;
	}

	if(core->config.user_commit & COMMIT_CHANNEL){
		if(!core->config.user.channel){
			if(core->config.active.network & IEEE_B)
				core->config.user.channel = geography_data[core->config.active.geography].bg[0].channel;
			else
				core->config.user.channel = geography_data[core->config.active.geography].a[0].channel;
		}

		if(!rt2x00_validate_config_channel(device, &core->config.active, core->config.user.channel)
		|| ieee80211_is_valid_channel(core->ieee80211, core->config.user.channel))
			core->config.active.channel = core->config.user.channel;
		else if(core->config.active.network & IEEE_B)
			core->config.active.channel = geography_data[core->config.active.geography].bg[0].channel;
		else
			core->config.active.channel = geography_data[core->config.active.geography].a[0].channel;
		if(rt2x00_is_52ghz_channel(core->config.active.channel))
			core->config.slot_time = 9;
		else
			core->config.slot_time = 20;
		core->config.update_flags |= UPDATE_CHANNEL | UPDATE_DURATION;
	}

	if(core->config.user_commit & COMMIT_BITRATE){
		if(!core->config.user.bitrate){
			if(core->config.active.network & IEEE_G)
				core->config.user.bitrate = IEEE80211_OFDM_RATE_54MB;
			else
				core->config.user.bitrate = IEEE80211_CCK_RATE_11MB;
		}

		if(!rt2x00_validate_config_bitrate(device, &core->config.active, core->config.user.bitrate))
			core->config.active.bitrate = core->config.user.bitrate;
		else if(core->config.active.network & IEEE_B)
			core->config.active.bitrate = IEEE80211_CCK_RATE_11MB;
		else
			core->config.active.bitrate = IEEE80211_OFDM_RATE_54MB;
		core->config.update_flags |= UPDATE_BITRATE;
	}

	if(core->config.user_commit & COMMIT_ESSID){
		if(core->config.user.flags & CONFIG_FIX_ESSID)
			memcpy(&core->config.active.essid, &core->config.user.essid, IW_ESSID_MAX_SIZE);
	}

	if(core->config.user_commit & COMMIT_BSSID){
		if(core->config.user.flags & CONFIG_FIX_BSSID)
			memcpy(&core->config.active.bssid, &core->config.user.bssid, ETH_ALEN);
		core->config.update_flags |= UPDATE_BSSID;
	}

	if(core->config.user_commit & COMMIT_RETRY){
		core->config.active.long_retry = core->config.user.long_retry;
		core->config.active.short_retry = core->config.user.short_retry;
		core->config.update_flags |= UPDATE_RETRY;
	}

	if(core->config.user_commit & COMMIT_TXPOWER){
		core->config.active.txpower = core->config.user.txpower;
		core->config.update_flags |= UPDATE_TXPOWER;
	}

	if(core->config.user_commit & COMMIT_PREAMBLE){
		core->config.update_flags |= UPDATE_PREAMBLE;
	}

	if(core->config.user_commit & COMMIT_ANTENNA){
		core->config.active.antenna_flags = core->config.user.antenna_flags;
		core->config.update_flags |= UPDATE_ANTENNA;
	}

	if(core->config.user_commit & COMMIT_SENSITIVITY){
		core->config.active.sensitivity = core->config.user.sensitivity;
		core->config.update_flags |= UPDATE_SENSITIVITY;
	}

	if(core->config.user_commit & COMMIT_RTS_THRESHOLD){
		core->config.active.rts_threshold = core->config.user.rts_threshold;
		spin_lock(&core->ieee80211->lock);
		core->ieee80211->rts = core->config.active.rts_threshold;
		spin_unlock(&core->ieee80211->lock);
	}

	if(core->config.user_commit & COMMIT_FRAGMENTATION){
		core->config.active.fragmentation = core->config.user.fragmentation;
		spin_lock(&core->ieee80211->lock);
		core->ieee80211->fts = core->config.active.fragmentation;
		spin_unlock(&core->ieee80211->lock);
	}

	if(core->config.user_commit & COMMIT_PACKET_FILTER){
		core->config.update_flags |= UPDATE_PACKET_FILTER;
	}
#if WIRELESS_EXT > 17
	if(core->config.user_commit & COMMIT_AUTHENTICATION){
		core->config.active.wpa_version |= core->config.user.wpa_version;
		core->config.active.mgmt_auth |= core->config.user.mgmt_auth;
	}
#endif /* WIRELESS_EXT > 17 */

	core->config.active.flags = core->config.user.flags;

	core->config.user_commit = 0;

	up_write(&device->rt2x00_sem);
}

/*
 * Link up/down routines.
 */
#define LINK_TIME			( HZ )

static void
rt2x00_link_up_config(struct _rt2x00_device *device)
{
	struct _rt2x00_core		*core = rt2x00_core(device);
	s8				counter = 0x00;
	u8				rate = 0x00;
	u8				bitrate_min = 0x00;
	u8				bitrate_max = 0x00;

	down_write(&device->rt2x00_sem);

	bitrate_min = (core->config.active.network & IEEE_B) ? 0 : 4;
	bitrate_max = (core->config.active.network == IEEE_B) ? 4 : 12;

	if(core->config.active.iw_mode == IW_MODE_MONITOR){
		memset(&core->config.active.essid, 0x00, sizeof(core->config.active.essid));
		memset(&core->config.active.bssid, 0x00, sizeof(core->config.active.bssid));
		core->config.active.channel = core->config.user.channel;
		core->config.update_flags |= UPDATE_BSSID | UPDATE_CHANNEL;
		up_write(&device->rt2x00_sem);
		return;
	}

	if(!core->network){
		up_write(&device->rt2x00_sem);
		return;
	}

	if(!ieee80211_is_empty_essid(&core->network->ssid[0], core->network->ssid_len))
		memcpy(&core->config.active.essid, &core->network->ssid, core->network->ssid_len);
	else 
		memcpy(&core->config.active.essid, "<hidden>", sizeof("<hidden>"));

	memcpy(&core->config.active.bssid, &core->network->bssid, sizeof(core->config.active.bssid));

	core->config.active.channel = core->network->channel;

	if(rt2x00_is_52ghz_channel(core->config.active.channel))
		core->config.slot_time = 9;
	else
		core->config.slot_time = 20;

	for(counter = bitrate_max; counter >= bitrate_min; counter--){
		rate = capabilities.bitrate[counter];
		if(rate < core->config.user.bitrate
		&& !rt2x00_validate_config_bitrate(device, &core->config.active, rate)){
			core->config.active.bitrate = rate;
			break;
		}
	}

	core->config.update_flags |= UPDATE_BSSID | UPDATE_CHANNEL | UPDATE_DURATION;

	up_write(&device->rt2x00_sem);
}

static void
rt2x00_link_down_config(struct _rt2x00_device *device)
{
	struct _rt2x00_core		*core = rt2x00_core(device);

	down_write(&device->rt2x00_sem);

	if(!(core->config.active.flags & CONFIG_FIX_ESSID))
		memset(&core->config.active.essid, 0x00, sizeof(core->config.active.essid));

	if(!(core->config.active.flags & CONFIG_FIX_BSSID))
		memset(&core->config.active.bssid, 0x00, sizeof(core->config.active.bssid));

	up_write(&device->rt2x00_sem);
}

static void
rt2x00_link_periodic_config(struct _rt2x00_device *device)
{
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_write(&device->rt2x00_sem);

	if(device->net_dev->flags & IFF_PROMISC){
		if(!(core->config.user.flags & CONFIG_ACCEPT_PROMISC))
			core->config.user_commit |= COMMIT_PACKET_FILTER;
		core->config.user.flags |= CONFIG_ACCEPT_PROMISC;
	}else{
		if(core->config.user.flags & CONFIG_ACCEPT_PROMISC)
			core->config.user_commit |= COMMIT_PACKET_FILTER;
		core->config.user.flags &= ~CONFIG_ACCEPT_PROMISC;
	}

	up_write(&device->rt2x00_sem);
}

static void
rt2x00_connect(struct _rt2x00_device *device, struct ieee80211_network *network, u8 mode)
{
	struct _rt2x00_core		*core = rt2x00_core(device);

	down_write(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);
	core->ieee80211->iw_mode = mode;
	core->config.active.iw_mode = mode;
	spin_unlock(&core->ieee80211->lock);
	up_write(&device->rt2x00_sem);

	core->network = network;

	rt2x00_link_up_config(device);
	rt2x00_link_periodic_config(device);

	rt2x00_update_config(device, core->config.update_flags);
}

static void
rt2x00_disconnect(struct _rt2x00_device *device)
{
	struct _rt2x00_core		*core = rt2x00_core(device);
	enum ieee80211_state		state = IEEE80211_UNINITIALIZED;

	spin_lock(&core->ieee80211->lock);
	state = core->ieee80211->state;
	spin_unlock(&core->ieee80211->lock);

	if(core->network){
		if(state == IEEE80211_ASSOCIATED)
			rt2x00_mgmt_snd_deassociation(device, &core->network->bssid[0], WLAN_REASON_DISASSOC_STA_HAS_LEFT);
		else if(state == IEEE80211_AUTHENTICATED)
			rt2x00_mgmt_snd_deauthentication(device, &core->network->bssid[0], WLAN_REASON_DEAUTH_LEAVING);
	}

	down_write(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);
	core->ieee80211->iw_mode = 0;
	core->config.active.iw_mode = 0;
	spin_unlock(&core->ieee80211->lock);
	up_write(&device->rt2x00_sem);

	core->network = NULL;

	rt2x00_link_down_config(device);
}

static void
rt2x00_link_up(struct _rt2x00_device *device)
{
	struct _rt2x00_core		*core = rt2x00_core(device);
	u8				mode = 0x00;
	u8				channel = 0x00;

	NOTICE("Link up.\n");

	down_read(&device->rt2x00_sem);
	mode = core->config.active.iw_mode;
	channel = (core->config.active.flags & CONFIG_FIX_CHANNEL);
	up_read(&device->rt2x00_sem);

	if(mode == IW_MODE_AUTO || mode == IW_MODE_ADHOC || mode == IW_MODE_INFRA){
		rt2x00_mgmt_snd_beacon(device);
		rt2x00_start_scan(device, (channel) ? 0 : IW_SCAN_ALL_FREQ);
	}else if(mode == IW_MODE_MONITOR){
		NOTICE("Enabling monitor mode.\n");
		down_write(&device->rt2x00_sem);
		device->net_dev->type = ARPHRD_IEEE80211;
		dev_set_promiscuity(device->net_dev, 1);
		up_write(&device->rt2x00_sem);

		rt2x00_connect(device, NULL, IW_MODE_MONITOR);
	}

	queue_delayed_work(core->workqueue, &core->link_tuner, LINK_TIME);
}

static void
rt2x00_link_down(struct _rt2x00_device *device)
{
	struct _rt2x00_core		*core = rt2x00_core(device);
	u8				mode = 0x00;

	down_read(&device->rt2x00_sem);
	mode = core->config.active.iw_mode;
	up_read(&device->rt2x00_sem);

	if(mode == IW_MODE_AUTO || mode == IW_MODE_ADHOC || mode == IW_MODE_INFRA)
		rt2x00_disconnect(device);

	if(mode == IW_MODE_MONITOR){
		down_write(&device->rt2x00_sem);
		device->net_dev->type = ARPHRD_ETHER;
		dev_set_promiscuity(device->net_dev, -1);
		up_write(&device->rt2x00_sem);

		rt2x00_disconnect(device);
	}

	NOTICE("Link down.\n");
}

static void
rt2x00_link_periodic(void *data)
{
	struct _rt2x00_device		*device = data;
	struct _rt2x00_core		*core = rt2x00_core(device);
	struct ieee80211_network	*network = NULL;
	enum ieee80211_state		state = IEEE80211_UNINITIALIZED;
	u8				iw_mode = 0x00;
	u8				channel = 0x00;
	u8				valid_entry = 0x00;
	u8				expired = 0x00;
	u8				connect = 0x00;

	if(test_bit(DEVICE_SCANNING, &device->flags))
		goto exit;

	down_read(&device->rt2x00_sem);
	iw_mode = core->config.active.iw_mode;
	channel = (core->config.active.flags & CONFIG_FIX_CHANNEL);
	expired = time_after(jiffies, core->connect.timestamp + CONNECT_TIMEOUT);
	connect = core->connect.status;
	up_read(&device->rt2x00_sem);

	if(iw_mode == IW_MODE_MONITOR){
		rt2x00_link_periodic_config(device);
		rt2x00_update_config(device, core->config.update_flags);
		goto exit;
	}

	spin_lock(&core->ieee80211->lock);
	state = core->ieee80211->state;
	if(state == IEEE80211_INITIALIZED)
		network = rt2x00_search_scan(device, iw_mode);
	else if(core->network)
		valid_entry = rt2x00_validate_scan_entry(core->ieee80211, &core->config, core->network);
	spin_unlock(&core->ieee80211->lock);

	if(state == IEEE80211_INITIALIZED){
		if(!network && (iw_mode == IW_MODE_AUTO || iw_mode == IW_MODE_ADHOC))
			rt2x00_connect(device, NULL, IW_MODE_ADHOC);
		else{
			rt2x00_connect(device, network, network->mode);
			rt2x00_mgmt_snd_authentication(device, &network->bssid[0]);
		}
	}else if(valid_entry){
		if((connect == CONNECT_AUTHENTICATE || connect == CONNECT_HANDSHAKE) && expired)
			rt2x00_mgmt_snd_authentication(device, &core->network->bssid[0]);
		else if(connect == CONNECT_ASSOCIATE && expired)
			rt2x00_mgmt_snd_association(device, &core->network->bssid[0]);
		else if(state == IEEE80211_ASSOCIATED){
			rt2x00_link_periodic_config(device);
			rt2x00_update_config(device, core->config.update_flags);
		}
	}else{
		NOTICE("Beacon lost, rescanning.\n");
		rt2x00_disconnect(device);
		rt2x00_start_scan(device, (channel) ? 0 : IW_SCAN_ALL_FREQ);
		goto exit;
	}

exit:
	queue_delayed_work(core->workqueue, &core->link_tuner, LINK_TIME);
}

/*
 * Radio control.
 */
static void
rt2x00_radio_on(struct _rt2x00_device *device)
{
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(!test_bit(DEVICE_ENABLED, &device->flags)){
		ERROR("Device not enabled.\n");
		return;
	}

	if(test_and_set_bit(DEVICE_RADIO_ON, &device->flags)){
		WARNING("Radio already on.\n");
		return;
	}

	rt2x00_enable_ring(&core->rx);
	rt2x00_enable_ring(&core->tx);
	rt2x00_enable_ring(&core->atim);
	rt2x00_enable_ring(&core->prio);
	rt2x00_enable_ring(&core->beacon);

	/*
	 * Start net interface tx /rx.
	 */
	down_write(&device->rt2x00_sem);
	netif_carrier_on(device->net_dev);
	netif_start_queue(device->net_dev);
	up_write(&device->rt2x00_sem);

	rt2x00_activate_config(device);
	rt2x00_update_config(device, UPDATE_ALL_CONFIG);
	device->handler->dev_radio_on(device);
	rt2x00_link_up(device);
}

static void
rt2x00_radio_off(struct _rt2x00_device *device)
{
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(!test_and_clear_bit(DEVICE_RADIO_ON, &device->flags)){
		WARNING("Radio already off.\n");
		return;
	}

	/*
	 * Stop net interface tx /rx.
	 */
	netif_stop_queue(device->net_dev);
	netif_carrier_off(device->net_dev);

	rt2x00_disable_ring(&core->rx);
	rt2x00_disable_ring(&core->tx);
	rt2x00_disable_ring(&core->atim);
	rt2x00_disable_ring(&core->prio);
	rt2x00_disable_ring(&core->beacon);

	cancel_delayed_work(&core->scan_tuner);
	cancel_delayed_work(&core->link_tuner);

	rt2x00_stop_scan(device);
	rt2x00_link_down(device);

	device->handler->dev_radio_off(device);
}

/*
 * IOCTL routines.
 */
static int
rt2x00_ioctl_commit(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u8			reset = 0x00;
	u8			mode = 0x00;

	down_read(&device->rt2x00_sem);
	reset = core->config.user_commit;
	mode = core->config.active.iw_mode;
	up_read(&device->rt2x00_sem);

	if(test_bit(DEVICE_RADIO_ON, &device->flags) && reset && mode != IW_MODE_MONITOR){
		INFO("SIOCSIWCOMMIT: Connection will be reset.\n");
		rt2x00_radio_off(device);
		rt2x00_radio_on(device);
	}else{
		INFO("SIOCSIWCOMMIT: Pending configuration will be loaded.\n");
		rt2x00_activate_config(device);
		rt2x00_update_config(device, core->config.update_flags);
	}

	return 0;
}

static int
rt2x00_ioctl_get_name(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	char 			*ieee80211_modes[] = {"?", "a", "b", "ab", "g", "ag", "bg", "abg"};

	down_read(&device->rt2x00_sem);

	snprintf(wrqu->name, sizeof(wrqu->name), "IEEE 802.11%s", ieee80211_modes[core->config.active.network]);

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_frequency(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u16			channel = 0x00;

	if(!((wrqu->freq.m > 0) && (wrqu->freq.m <= 1000)))
		channel = ieee80211_freq_to_channel(core->ieee80211, wrqu->freq.m);
	else
		channel = wrqu->freq.m;

	down_write(&device->rt2x00_sem);

	if(rt2x00_validate_config_channel(device, &core->config.user, channel)){
		up_write(&device->rt2x00_sem);
		return -EINVAL;
	}

	core->config.user.channel = channel;
	core->config.user_commit |= COMMIT_CHANNEL;
	core->config.user.flags |= CONFIG_FIX_CHANNEL;

	up_write(&device->rt2x00_sem);

	INFO("SIOCSIWFREQ: Channel = %d.\n", channel);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_frequency(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u8			channel = 0x00;

	down_read(&device->rt2x00_sem);

	if(test_bit(DEVICE_SCANNING, &device->flags))
		channel = core->config.user.channel;
	else
		channel = core->config.active.channel;

	up_read(&device->rt2x00_sem);

	wrqu->freq.m = rt2x00_channel_to_freq(channel) * 100000;
	wrqu->freq.e = 1;
	wrqu->freq.i = channel;

	return 0;
}

static int 
rt2x00_ioctl_set_mode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(wrqu->mode == IW_MODE_REPEAT
	|| wrqu->mode == IW_MODE_MASTER
	|| wrqu->mode == IW_MODE_SECOND)
		return -ENOTSUPP;

	down_write(&device->rt2x00_sem);

	if(rt2x00_validate_config_mode(device, &core->config.user, wrqu->mode)){
		up_write(&device->rt2x00_sem);
		return -EINVAL;
	}

	core->config.user.iw_mode = wrqu->mode;
	core->config.user_commit |= COMMIT_MODE;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_mode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	if(core->config.active.iw_mode)
		wrqu->mode = core->config.active.iw_mode;
	else
		wrqu->mode = core->config.user.iw_mode;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_sens(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_write(&device->rt2x00_sem);

	if(wrqu->sens.value < 0)	/* dBm. */
		core->config.user.sensitivity = wrqu->sens.value + 120;
	else				/* RSSI. */
		core->config.user.sensitivity = wrqu->sens.value;

	core->config.user_commit |= COMMIT_SENSITIVITY;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_sens(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->sens.value = core->config.active.sensitivity;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_get_range(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device		*device = ieee80211_priv(net_dev);
	struct _rt2x00_core		*core = rt2x00_core(device);
	struct iw_range			*range = (struct iw_range *)extra;
	const struct ieee80211_geo	*geo = ieee80211_get_geo(core->ieee80211);
	u8				bitrate_min = 0x00;
	u8				bitrate_max = 0x00;
	u8				network = 0x00;
	u8				counter = 0x00;

	memset(range, 0x00, sizeof(struct iw_range));

	down_read(&device->rt2x00_sem);
	bitrate_min = (core->config.active.network & IEEE_B) ? 0 : 4;
	bitrate_max = (core->config.active.network == IEEE_B) ? 4 : 12;
	network = core->config.active.network;
	up_read(&device->rt2x00_sem);

	range->throughput = capabilities.bitrate[bitrate_max - 1] * 500000;

	range->event_capa[0] = IW_EVENT_CAPA_K_0
				| IW_EVENT_CAPA_MASK(SIOCSIWSPY)
				| IW_EVENT_CAPA_MASK(SIOCSIWTHRSPY);
	range->event_capa[1] = IW_EVENT_CAPA_K_1
				| IW_EVENT_CAPA_MASK(SIOCSIWAP)
				| IW_EVENT_CAPA_MASK(SIOCSIWSCAN)
				| IW_EVENT_CAPA_MASK(SIOCSIWNICKN)
				| IW_EVENT_CAPA_MASK(SIOCSIWRATE);

	range->event_capa[2] = IW_EVENT_CAPA_MASK(SIOCSIWRTS)
				| IW_EVENT_CAPA_MASK(SIOCSIWFRAG)
				| IW_EVENT_CAPA_MASK(SIOCSIWTXPOW)
				| IW_EVENT_CAPA_MASK(SIOCSIWRETRY);
#if WIRELESS_EXT > 17
	range->event_capa[3] = IW_EVENT_CAPA_MASK(SIOCSIWAUTH)
				| IW_EVENT_CAPA_MASK(SIOCSIWENCODEEXT);
#endif /* WIRELESS_EXT > 17 */

	range->num_bitrates = bitrate_max;
	for(counter = 0x00; counter < bitrate_max; counter++)
		range->bitrate[counter] = capabilities.bitrate[counter] * 500000;

	range->min_rts	= 0;
	range->max_rts	= MAX_FRAG_THRESHOLD;

	range->min_frag = MIN_FRAG_THRESHOLD;
	range->max_frag = MAX_FRAG_THRESHOLD;

	range->encoding_size[0] = 5;	/* 64bit WEP. */
	range->encoding_size[1] = 13;	/* 128bit WEP. */
	range->num_encoding_sizes = 2;
	range->max_encoding_tokens = 4;

	range->txpower_capa = IW_TXPOW_RELATIVE;
	range->num_txpower = ARRAY_SIZE(capabilities.txpower);
	for(counter = 0x00; counter < range->num_txpower; counter++)
		range->txpower[counter] = capabilities.txpower[counter];

	range->we_version_compiled = WIRELESS_EXT;
	range->we_version_source = 17;

	range->retry_capa = IW_RETRY_LIMIT;
	range->retry_flags = IW_RETRY_LIMIT | IW_RETRY_MIN | IW_RETRY_MAX;
	range->min_retry = 0;
	range->max_retry = 255;

	spin_lock(&core->ieee80211->lock);

	range->num_channels = 0;
	if((network & IEEE_B))
		range->num_channels += geo->bg_channels;
	if(network & IEEE_A)
		range->num_channels += geo->a_channels;

	if(range->num_channels > IW_MAX_FREQUENCIES)
		range->num_channels = IW_MAX_FREQUENCIES;

	range->num_frequency = range->num_channels;

	for(counter = 0x00; counter < range->num_frequency; counter++) {
		if((network & IEEE_B) && counter < geo->bg_channels){
			range->freq[counter].i = geo->bg[counter].channel;
			range->freq[counter].m = geo->bg[counter].freq * 100000;
			range->freq[counter].e = 1;
		}else if((network & IEEE_A) && (counter - geo->bg_channels) < geo->a_channels){
			range->freq[counter].i = geo->a[counter - geo->bg_channels].channel;
			range->freq[counter].m = geo->a[counter - geo->bg_channels].freq * 100000;
			range->freq[counter].e = 1;
		}
	}

	spin_unlock(&core->ieee80211->lock);

#if WIRELESS_EXT > 17
	range->enc_capa = IW_ENC_CAPA_WPA | IW_ENC_CAPA_WPA2 | IW_ENC_CAPA_CIPHER_TKIP | IW_ENC_CAPA_CIPHER_CCMP;
#endif /* WIRELESS_EXT > 17 */

	wrqu->data.length = sizeof(struct iw_range);

	return 0;
}

static int
rt2x00_ioctl_set_wap(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(wrqu->ap_addr.sa_family != ARPHRD_ETHER)
		return -EINVAL;

	down_write(&device->rt2x00_sem);

	/*
	 * When the address is not valid, because it is, for example, a broadcast address,
	 * we must set the essid to 'any' instead.
	 */
	if(is_valid_ether_addr(&wrqu->ap_addr.sa_data[0])){
		memcpy(&core->config.user.bssid, wrqu->ap_addr.sa_data, ETH_ALEN);

		core->config.user_commit |= COMMIT_BSSID;
		core->config.user.flags |= CONFIG_FIX_BSSID;
	} else {
		core->config.user_commit |= COMMIT_BSSID;
		core->config.user.flags &= ~CONFIG_FIX_BSSID;
	}

		INFO("SIOCSIWAP: Bssid = " ADDR_STRING ".\n", ADDR_PRINT(core->config.user.bssid));

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_wap(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->ap_addr.sa_family = ARPHRD_ETHER;

	if(core->config.active.iw_mode == IW_MODE_ADHOC
	|| core->config.active.iw_mode == IW_MODE_INFRA)
		memcpy(wrqu->ap_addr.sa_data, core->config.active.bssid, ETH_ALEN);
	else
		memset(wrqu->ap_addr.sa_data, 0x00, ETH_ALEN);

	up_read(&device->rt2x00_sem);

	return 0;
}

#if WIRELESS_EXT > 17
static int
rt2x00_ioctl_set_mlme(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct iw_mlme		*mlme = (struct iw_mlme*)wrqu->data.pointer;

	if(!test_bit(DEVICE_RADIO_ON, &device->flags))
		return -ENOTCONN;

	if(mlme->cmd == IW_MLME_DEAUTH)
		rt2x00_mgmt_snd_deauthentication(device, &mlme->addr.sa_data[0], mlme->reason_code);
	if(mlme->cmd == IW_MLME_DISASSOC)
		rt2x00_mgmt_snd_deassociation(device, &mlme->addr.sa_data[0], mlme->reason_code);

	return 0;
}
#endif /* WIRELESS_EXT > 17 */

static int
rt2x00_ioctl_set_scan(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = rt2x00_device(net_dev);

	if(!test_bit(DEVICE_RADIO_ON, &device->flags))
		return -ENOTCONN;

	if(test_bit(DEVICE_SCANNING, &device->flags)){
		NOTICE("Device already busy scanning, aborting current scan...\n");
		rt2x00_stop_scan(device);
	}

	if(wrqu->param.flags)
		rt2x00_start_scan(device, wrqu->param.flags);
	else
		rt2x00_start_scan(device, IW_SCAN_ALL_FREQ);

	return 0;
}

static int
rt2x00_ioctl_get_scan(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device		*device = rt2x00_device(net_dev);
	struct _rt2x00_core		*core = rt2x00_core(device);
	const struct ieee80211_geo	*geo = ieee80211_get_geo(core->ieee80211);
	u8				counter = 0x00;
	u8				max = 0x00;

	spin_lock(&core->ieee80211->lock);
	max = geo->bg_channels + geo->a_channels + 2;
	spin_unlock(&core->ieee80211->lock);

	for(; counter < max; counter++){
		if(!test_bit(DEVICE_SCANNING, &device->flags))
			return ieee80211_wx_get_scan(core->ieee80211, info, wrqu, extra);
		msleep(SCAN_TIME);
	}

	ERROR("Timeout waiting for adapter scan results.\n");

	return -EBUSY;
}

static int
rt2x00_ioctl_set_essid(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_write(&device->rt2x00_sem);

	if(wrqu->essid.flags){
		if(wrqu->essid.length > IW_ESSID_MAX_SIZE) {
			up_write(&device->rt2x00_sem);
			return -E2BIG;
		}

		if(wrqu->essid.length > 0){
			memcpy(&core->config.user.essid, extra, wrqu->essid.length);
			core->config.user.essid[wrqu->essid.length] = '\0';
		}

		core->config.user.flags |= CONFIG_FIX_ESSID;
	}else
		core->config.user.flags &= ~CONFIG_FIX_ESSID;

	core->config.user_commit |= COMMIT_ESSID;

	up_write(&device->rt2x00_sem);

	if (wrqu->essid.flags)
		INFO("SIOCSIWESSID: Essid = %s length = %d.\n", extra, wrqu->essid.length);
	else
		INFO("SIOCSIWESSID: Essid = ANY\n");

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_essid(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->essid.length = strnlen(core->config.active.essid, IW_ESSID_MAX_SIZE);

	if(wrqu->essid.length){
		memcpy(extra, core->config.active.essid, wrqu->essid.length);
		wrqu->essid.flags = 1;
	}else
		wrqu->essid.flags = 0;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_nickname(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	if (wrqu->data.length > IW_ESSID_MAX_SIZE)
		return -E2BIG;

	down_write(&device->rt2x00_sem);

	memset(&core->config.nickn, 0x00, IW_ESSID_MAX_SIZE);

	memcpy(&core->config.nickn, extra, wrqu->data.length);

	core->config.nickn[wrqu->data.length] = '\0';

	up_write(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_get_nickname(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->data.length = strnlen(core->config.nickn, IW_ESSID_MAX_SIZE);
	memcpy(extra, core->config.nickn, wrqu->data.length);

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_bitrate(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u8			rate = wrqu->bitrate.value / 500000;

	down_write(&device->rt2x00_sem);

	if(wrqu->bitrate.value == -1)
		rate = 0;
	else{
		if(rt2x00_validate_config_bitrate(device, &core->config.user, rate)){
			up_write(&device->rt2x00_sem);
			return -EINVAL;
		}
	}

	core->config.user.bitrate = rate;

	if(wrqu->bitrate.fixed)
		core->config.user.flags |= CONFIG_FIX_BITRATE;
	else
		core->config.user.flags &= ~CONFIG_FIX_BITRATE;

	core->config.user_commit |= COMMIT_BITRATE;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_bitrate(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->bitrate.value = core->config.active.bitrate * 500000;
	wrqu->bitrate.fixed = core->config.active.flags & CONFIG_FIX_BITRATE;
	wrqu->bitrate.disabled = 0;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_rts_threshold(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(wrqu->rts.value != -1 && wrqu->rts.value > MAX_FRAG_THRESHOLD)
		return -EINVAL;

	down_write(&device->rt2x00_sem);

	if(wrqu->rts.disabled)
		core->config.user.flags &= ~CONFIG_ENABLED_RTS;
	else{
		core->config.user.flags |= CONFIG_ENABLED_RTS;
		core->config.user.rts_threshold = wrqu->rts.value;
	}

	if(wrqu->rts.fixed)
		core->config.user.flags |= CONFIG_FIX_RTS;
	else
		core->config.user.flags &= ~CONFIG_FIX_RTS;

	core->config.user_commit |= COMMIT_RTS_THRESHOLD;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_rts_threshold(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->rts.value = core->config.active.rts_threshold;
	wrqu->rts.disabled = !(core->config.active.flags & CONFIG_ENABLED_RTS);
	wrqu->rts.fixed = core->config.active.flags & CONFIG_FIX_RTS;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_fragmentation(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(wrqu->frag.value != -1
	&& (MIN_FRAG_THRESHOLD > wrqu->frag.value || wrqu->frag.value > MAX_FRAG_THRESHOLD))
		return -EINVAL;

	down_write(&device->rt2x00_sem);

	if(wrqu->frag.disabled){
		core->config.user.flags &= ~CONFIG_ENABLED_FRAG;
		core->config.user.fragmentation = DEFAULT_FTS;
	}else{
		core->config.user.flags |= CONFIG_ENABLED_FRAG;
		core->config.user.fragmentation = wrqu->frag.value;
	}

	if(wrqu->frag.fixed)
		core->config.user.flags |= CONFIG_FIX_FRAG;
	else
		core->config.user.flags &= ~CONFIG_FIX_FRAG;

	core->config.user_commit |= COMMIT_FRAGMENTATION;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_fragmentation(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->frag.value = core->config.active.fragmentation;
	wrqu->frag.disabled = !(core->config.active.flags & CONFIG_ENABLED_FRAG);
	wrqu->frag.fixed = core->config.active.flags & CONFIG_FIX_FRAG;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_txpower(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u8			txpower = 0x00;
	u8			counter = 0x00;
	u8			max = 0x00;

	if(test_bit(DEVICE_ENABLED, &device->flags)){
		if(test_bit(DEVICE_RADIO_ON, &device->flags)
		&& (wrqu->txpower.disabled || wrqu->txpower.value == 0)){
			INFO("SIOCSIWTXPOW: Radio off.\n");
			rt2x00_radio_off(device);
		}else if(!test_bit(DEVICE_RADIO_ON, &device->flags)){
			INFO("SIOCSIWTXPOW: Radio on.\n");
			rt2x00_radio_on(device);
		}
	}

	if(wrqu->txpower.flags & IW_TXPOW_DBM
	|| wrqu->txpower.flags & IW_TXPOW_MWATT)
		return -ENOTSUPP;

	max = ARRAY_SIZE(capabilities.txpower);

	for(; counter < max; counter++){
		txpower = capabilities.txpower[counter];
		if(txpower >= wrqu->txpower.value)
			break;
	}

	down_write(&device->rt2x00_sem);

	core->config.user.txpower = txpower;

	if(wrqu->txpower.fixed)
		core->config.user.flags |= CONFIG_FIX_TXPOWER;
	else
		core->config.user.flags &= ~CONFIG_FIX_TXPOWER;

	core->config.user_commit |= COMMIT_TXPOWER;

	up_write(&device->rt2x00_sem);

	INFO("SIOCSIWTXPOW: Tx power = %d%s.\n", txpower, "%");

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_txpower(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu, 
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	wrqu->txpower.value = core->config.active.txpower;
	wrqu->txpower.disabled = !test_bit(DEVICE_RADIO_ON, &device->flags);
	wrqu->txpower.fixed = core->config.active.flags & CONFIG_FIX_TXPOWER;
	wrqu->txpower.flags |= IW_TXPOW_RELATIVE;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_retry(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(!(wrqu->retry.flags & IW_RETRY_LIMIT))
		return -ENOTSUPP;

	down_write(&device->rt2x00_sem);

	if(wrqu->retry.flags & IW_RETRY_MAX){
		core->config.user.long_retry = wrqu->retry.value;
		core->config.user.flags |= CONFIG_ENABLED_LONG_RETRY;
	}else if(wrqu->retry.flags & IW_RETRY_MIN){
		core->config.user.short_retry = wrqu->retry.value;
		core->config.user.flags &= ~CONFIG_ENABLED_LONG_RETRY;
	}else{
		core->config.user.short_retry = wrqu->retry.value;
		core->config.user.long_retry = wrqu->retry.value;
		core->config.user.flags |= CONFIG_ENABLED_LONG_RETRY;
	}

	core->config.user_commit |= COMMIT_RETRY;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_retry(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(!(wrqu->retry.flags & IW_RETRY_LIMIT))
		return -ENOTSUPP;

	down_read(&device->rt2x00_sem);

	if(wrqu->retry.flags & IW_RETRY_MAX)
		wrqu->retry.value = core->config.active.long_retry;
	else if(wrqu->retry.flags & IW_RETRY_MIN)
		wrqu->retry.value = core->config.active.short_retry;
	else if(core->config.active.flags & CONFIG_ENABLED_LONG_RETRY){
		wrqu->retry.value = core->config.active.long_retry;
		wrqu->retry.flags |= IW_RETRY_MAX;
	}else{
		wrqu->retry.value = core->config.active.short_retry;
		wrqu->retry.flags |= IW_RETRY_MIN;
	}

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_encode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	int			status = 0x00000000;

	status = ieee80211_wx_set_encode(netdev_priv(net_dev), info, wrqu, extra);
	if(status)
		return status;

	down_write(&device->rt2x00_sem);

	core->config.user_commit |= COMMIT_ENCRYPTION;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_encode(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	return ieee80211_wx_get_encode(netdev_priv(net_dev), info, wrqu, extra);
}

#if WIRELESS_EXT > 17
static int
rt2x00_ioctl_set_auth(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	int			status = -EIWCOMMIT;

	down_write(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	if(wrqu->param.flags == IW_AUTH_WPA_VERSION)
		core->config.user.wpa_version = wrqu->param.value;

	if(wrqu->param.flags == IW_AUTH_KEY_MGMT){
		if(wrqu->param.value == IW_AUTH_KEY_MGMT_802_1X){
			ERROR("SIOCSIWAUTH: IW_AUTH_KEY_MGMT_802_1X not supported.\n");
			status = -ENOTSUPP;
		} else
			core->config.user.mgmt_auth = wrqu->param.value;
	}

	if(wrqu->param.flags == IW_AUTH_TKIP_COUNTERMEASURES)
		core->ieee80211->tkip_countermeasures = wrqu->param.value;

	if(wrqu->param.flags == IW_AUTH_DROP_UNENCRYPTED)
		core->ieee80211->drop_unencrypted = wrqu->param.value;

	if(wrqu->param.flags == IW_AUTH_80211_AUTH_ALG){
		if(wrqu->param.value == IW_AUTH_ALG_LEAP){
			ERROR("SIOCSIWAUTH: IW_AUTH_ALG_LEAP not supported.\n");
			status = -ENOTSUPP;
		} else
			core->ieee80211->sec.auth_mode = wrqu->param.value;
	}

	if(wrqu->param.flags == IW_AUTH_WPA_ENABLED)
		core->ieee80211->wpa_enabled = wrqu->param.value;

	if(wrqu->param.flags == IW_AUTH_PRIVACY_INVOKED)
		core->ieee80211->privacy_invoked = wrqu->param.value;

	if(wrqu->param.flags == IW_AUTH_CIPHER_PAIRWISE){
		ERROR("SIOCSIWAUTH: IW_AUTH_CIPHER_PAIRWISE not supported.\n");
		status = -ENOTSUPP;
	}

	if(wrqu->param.flags == IW_AUTH_CIPHER_GROUP){
		ERROR("SIOCSIWAUTH: IW_AUTH_CIPHER_GROUP not supported.\n");
		status = -ENOTSUPP;
	}

	if(wrqu->param.flags == IW_AUTH_RX_UNENCRYPTED_EAPOL){
		ERROR("SIOCSIWAUTH: IW_AUTH_RX_UNENCRYPTED_EAPOL not supported.\n");
		status = -ENOTSUPP;
	}

	if(wrqu->param.flags == IW_AUTH_ROAMING_CONTROL){
		ERROR("SIOCSIWAUTH: IW_AUTH_ROAMING_CONTROL not supported.\n");
		status = -ENOTSUPP;
	}

	spin_unlock(&core->ieee80211->lock);

	core->config.user_commit |= COMMIT_AUTHENTICATION;

	up_write(&device->rt2x00_sem);

	return status;
}

static int
rt2x00_ioctl_get_auth(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	int			status = 0x00000000;

	down_read(&device->rt2x00_sem);
	spin_lock(&core->ieee80211->lock);

	if(wrqu->param.flags == IW_AUTH_WPA_VERSION)
		wrqu->param.value = core->config.user.wpa_version;

	if(wrqu->param.flags == IW_AUTH_KEY_MGMT)
		wrqu->param.value = core->config.user.mgmt_auth;

	if(wrqu->param.flags == IW_AUTH_TKIP_COUNTERMEASURES)
		wrqu->param.value = core->ieee80211->tkip_countermeasures;

	if(wrqu->param.flags == IW_AUTH_DROP_UNENCRYPTED)
		wrqu->param.value = core->ieee80211->drop_unencrypted;

	if(wrqu->param.flags == IW_AUTH_80211_AUTH_ALG)
		wrqu->param.value = core->ieee80211->sec.auth_mode;

	if(wrqu->param.flags == IW_AUTH_WPA_ENABLED)
		wrqu->param.value = core->ieee80211->wpa_enabled;

	if(wrqu->param.flags == IW_AUTH_PRIVACY_INVOKED)
		wrqu->param.value = core->ieee80211->privacy_invoked;

	if(wrqu->param.flags == IW_AUTH_CIPHER_PAIRWISE){
		ERROR("SIOCGIWAUTH: IW_AUTH_CIPHER_PAIRWISE not supported.\n");
		status = -ENOTSUPP;
	}

	if(wrqu->param.flags == IW_AUTH_CIPHER_GROUP){
		ERROR("SIOCGIWAUTH: IW_AUTH_CIPHER_GROUP not supported.\n");
		status = -ENOTSUPP;
	}

	if(wrqu->param.flags == IW_AUTH_RX_UNENCRYPTED_EAPOL){
		ERROR("SIOCGIWAUTH: IW_AUTH_RX_UNENCRYPTED_EAPOL not supported.\n");
		status = -ENOTSUPP;
	}

	if(wrqu->param.flags == IW_AUTH_ROAMING_CONTROL){
		ERROR("SIOCGIWAUTH: IW_AUTH_ROAMING_CONTROL not supported.\n");
		status = -ENOTSUPP;
	}

	spin_unlock(&core->ieee80211->lock);
	up_read(&device->rt2x00_sem);

	return status;
}

static int
rt2x00_ioctl_set_encodeext(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	int			res = 0x00000000;

	res = ieee80211_wx_set_encodeext(netdev_priv(net_dev), info, wrqu, extra);
	if(res)
		return res;

	down_write(&device->rt2x00_sem);

	core->config.user_commit |= COMMIT_ENCRYPTION;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_encodeext(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	return ieee80211_wx_get_encodeext(netdev_priv(net_dev), info, wrqu, extra);
}
#endif /* WIRELESS_EXT > 17 */

static int
rt2x00_ioctl_set_network(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u8			network = *extra;

	if((network & IEEE_A) && !test_bit(DEVICE_CAP_802_11A, &device->flags)){
		DEBUG("SIOCSIWPRIVNETWORK: Device does not support 802.11a.\n");
		return 0;
	}

	if((network & IEEE_G) && !test_bit(DEVICE_CAP_802_11G, &device->flags)){
		DEBUG("SIOCSIWPRIVNETWORK: Device does not support 802.11g.\n");
		return 0;
	}

	if(!(network & IEEE_MODE_MASK))
		network = 0;

	if(network & IEEE_G)
		network |= IEEE_B;

	if(network & IEEE_A && network & IEEE_B)
		network |= IEEE_G;

	down_write(&device->rt2x00_sem);

	core->config.user.network = network;
	core->config.user_commit |= COMMIT_NETWORK;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_network(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	*extra = core->config.active.network;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_preamble(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_write(&device->rt2x00_sem);

	if(*extra & IOCTL_PREAMBLE_SHORT)
		core->config.user.flags |= CONFIG_ENABLED_SHORT_PREAMBLE;
	else
		core->config.user.flags &= ~CONFIG_ENABLED_SHORT_PREAMBLE;

	core->config.user_commit |= COMMIT_PREAMBLE;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_preamble(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	if(core->config.active.flags & CONFIG_ENABLED_SHORT_PREAMBLE)
		*extra = IOCTL_PREAMBLE_SHORT;
	else
		*extra = IOCTL_PREAMBLE_LONG;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_antenna(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u8			antenna = *extra;

	if(!(antenna & IOCTL_ANTENNA_RX_A)
	|| !(antenna & IOCTL_ANTENNA_RX_B)
	|| !(antenna & IOCTL_ANTENNA_RX_DIV)){
		DEBUG("SIOCSIWPRIVANTENNA: Invalid rx-antenna value, switching rx_antenna to diversity.\n");
		antenna |= ANTENNA_RX_DIV;
	}
	if(!(antenna & IOCTL_ANTENNA_TX_A)
	|| !(antenna & IOCTL_ANTENNA_TX_B)
	|| !(antenna & IOCTL_ANTENNA_TX_DIV)){
		DEBUG("SIOCSIWPRIVANTENNA: Invalid tx_antenna value, switching tx_antenna to diversity.\n");
		antenna |= ANTENNA_TX_DIV;
	}

	down_write(&device->rt2x00_sem);

	core->config.user.antenna_flags = 0;

	if(antenna & IOCTL_ANTENNA_RX_A)
		core->config.user.antenna_flags |= ANTENNA_RX_A;
	else if(antenna & IOCTL_ANTENNA_RX_B)
		core->config.user.antenna_flags |= ANTENNA_RX_B;
	else
		core->config.user.antenna_flags |= ANTENNA_RX_DIV;

	if(antenna & IOCTL_ANTENNA_TX_A)
		core->config.user.antenna_flags |= ANTENNA_TX_A;
	else if(antenna & IOCTL_ANTENNA_TX_B)
		core->config.user.antenna_flags |= ANTENNA_TX_B;
	else
		core->config.user.antenna_flags |= ANTENNA_TX_DIV;

	core->config.user_commit |= COMMIT_ANTENNA;

	up_write(&device->rt2x00_sem);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_antenna(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	if(core->config.active.antenna_flags & ANTENNA_RX_A)
		*extra |= IOCTL_ANTENNA_TX_A;
	else if(core->config.active.antenna_flags & ANTENNA_RX_B)
		*extra |= IOCTL_ANTENNA_TX_B;
	else if(core->config.active.antenna_flags & ANTENNA_RX_DIV)
		*extra |= IOCTL_ANTENNA_TX_DIV;

	if(core->config.active.antenna_flags & ANTENNA_TX_A)
		*extra |= IOCTL_ANTENNA_RX_A;
	else if(core->config.active.antenna_flags & ANTENNA_TX_B)
		*extra |= IOCTL_ANTENNA_RX_B;
	else if(core->config.active.antenna_flags & ANTENNA_TX_DIV)
		*extra |= IOCTL_ANTENNA_RX_DIV;

	up_read(&device->rt2x00_sem);

	return 0;
}

static int
rt2x00_ioctl_set_geography(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);
	u8			geography = *extra;

	if(geography >= IOCTL_GEOGRAPHY_INVAL)
		return -EINVAL;

	down_write(&device->rt2x00_sem);

	core->config.user.geography = geography;
	core->config.user_commit |= COMMIT_GEOGRAPHY;

	up_write(&device->rt2x00_sem);

	INFO("SIOCSIWPRIVGEOGRAPHY: Geography set to %s.\n", geography_data[geography].name);

	return -EIWCOMMIT;
}

static int
rt2x00_ioctl_get_geography(
	struct net_device *net_dev,
	struct iw_request_info *info,
	union iwreq_data *wrqu,
	char *extra)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	down_read(&device->rt2x00_sem);

	*extra = core->config.active.geography;

	up_read(&device->rt2x00_sem);

	return 0;
}

static struct iw_statistics*
rt2x00_get_wireless_stats(struct net_device *net_dev)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	return	&core->wireless_stats;
}

static const iw_handler rt2x00_ioctl_handler[] = {
	(iw_handler) rt2x00_ioctl_commit,		/* SIOCSIWCOMMIT */
	(iw_handler) rt2x00_ioctl_get_name,		/* SIOCGIWNAME */
	(iw_handler) NULL,				/* SIOCSIWNWID */
	(iw_handler) NULL,				/* SIOCGIWNWID */
	(iw_handler) rt2x00_ioctl_set_frequency,	/* SIOCSIWFREQ */
	(iw_handler) rt2x00_ioctl_get_frequency,	/* SIOCGIWFREQ */
	(iw_handler) rt2x00_ioctl_set_mode,		/* SIOCSIWMODE */
	(iw_handler) rt2x00_ioctl_get_mode,		/* SIOCGIWMODE */
	(iw_handler) rt2x00_ioctl_set_sens,		/* SIOCSIWSENS */
	(iw_handler) rt2x00_ioctl_get_sens,		/* SIOCGIWSENS */
	(iw_handler) NULL, /* Unused */			/* SIOCSIWRANGE */
	(iw_handler) rt2x00_ioctl_get_range,		/* SIOCGIWRANGE */
	(iw_handler) NULL, /* Unused */			/* SIOCSIWPRIV */
	(iw_handler) NULL, /* Provided by kernel */	/* SIOCGIWPRIV */
	(iw_handler) NULL, /* Unused */			/* SIOCSIWSTATS */
	(iw_handler) NULL, /* Provided by kernel */	/* SIOCGIWSTATS */
	(iw_handler) iw_handler_set_spy,		/* SIOCSIWSPY */
	(iw_handler) iw_handler_get_spy,		/* SIOCGIWSPY */
	(iw_handler) iw_handler_set_thrspy,		/* SIOCSIWTHRSPY */
	(iw_handler) iw_handler_get_thrspy,		/* SIOCGIWTHRSPY */
	(iw_handler) rt2x00_ioctl_set_wap,		/* SIOCSIWAP */
	(iw_handler) rt2x00_ioctl_get_wap,		/* SIOCGIWAP */
#if WIRELESS_EXT > 17
	(iw_handler) rt2x00_ioctl_set_mlme,		/* SIOCSIWMLME */
#else /* WIRELESS_EXT > 17 */
	(iw_handler) NULL,				/* SIOCSIWAPLIST */
#endif /* WIRELESS_EXT > 17 */
	(iw_handler) NULL, /* Deprecated */		/* SIOCGIWAPLIST */
	(iw_handler) rt2x00_ioctl_set_scan,		/* SIOCSIWSCAN */
	(iw_handler) rt2x00_ioctl_get_scan,		/* SIOCGIWSCAN */
	(iw_handler) rt2x00_ioctl_set_essid,		/* SIOCSIWESSID */
	(iw_handler) rt2x00_ioctl_get_essid,		/* SIOCGIWESSID */
	(iw_handler) rt2x00_ioctl_set_nickname,		/* SIOCSIWNICKN */
	(iw_handler) rt2x00_ioctl_get_nickname,		/* SIOCGIWNICKN */
	(iw_handler) NULL,				/* -- hole -- */
	(iw_handler) NULL,				/* -- hole -- */
	(iw_handler) rt2x00_ioctl_set_bitrate,		/* SIOCSIWRATE */
	(iw_handler) rt2x00_ioctl_get_bitrate,		/* SIOCGIWRATE */
	(iw_handler) rt2x00_ioctl_set_rts_threshold,	/* SIOCSIWRTS */
	(iw_handler) rt2x00_ioctl_get_rts_threshold,	/* SIOCGIWRTS */
	(iw_handler) rt2x00_ioctl_set_fragmentation,	/* SIOCSIWFRAG */
	(iw_handler) rt2x00_ioctl_get_fragmentation,	/* SIOCGIWFRAG */
	(iw_handler) rt2x00_ioctl_set_txpower,		/* SIOCSIWTXPOW */
	(iw_handler) rt2x00_ioctl_get_txpower,		/* SIOCGIWTXPOW */
	(iw_handler) rt2x00_ioctl_set_retry,		/* SIOCSIWRETRY */
	(iw_handler) rt2x00_ioctl_get_retry,		/* SIOCGIWRETRY */
	(iw_handler) rt2x00_ioctl_set_encode,		/* SIOCSIWENCODE */
	(iw_handler) rt2x00_ioctl_get_encode,		/* SIOCGIWENCODE */
	(iw_handler) NULL,				/* SIOCSIWPOWER */
	(iw_handler) NULL,				/* SIOCGIWPOWER */
#if WIRELESS_EXT > 17
	(iw_handler) NULL,				/* -- hole -- */
	(iw_handler) NULL,				/* -- hole -- */
	(iw_handler) NULL,				/* SIOCSIWGENIE */
	(iw_handler) NULL,				/* SIOCGIWGENIE */
	(iw_handler) rt2x00_ioctl_set_auth,		/* SIOCSIWAUTH */
	(iw_handler) rt2x00_ioctl_get_auth,		/* SIOCGIWAUTH */
	(iw_handler) rt2x00_ioctl_set_encodeext,	/* SIOCSIWENCODEEXT */
	(iw_handler) rt2x00_ioctl_get_encodeext,	/* SIOCGIWENCODEEXT */
#endif /* WIRELESS_EXT > 17 */
};

static const iw_handler rt2x00_ioctl_priv_handler[] = {
	(iw_handler) rt2x00_ioctl_set_network,			/* SIOCSIWPRIVNETWORK */
	(iw_handler) rt2x00_ioctl_get_network,			/* SIOCGIWPRIVNETWORK */
	(iw_handler) rt2x00_ioctl_set_preamble,			/* SIOCSIWPRIVPREAMBLE */
	(iw_handler) rt2x00_ioctl_get_preamble,			/* SIOCGIWPRIVPREAMBLE */
	(iw_handler) rt2x00_ioctl_set_antenna,			/* SIOCSIWPRIVANTENNA */
	(iw_handler) rt2x00_ioctl_get_antenna,			/* SIOCGIWPRIVANTENNA */
	(iw_handler) rt2x00_ioctl_set_geography,		/* SIOCSIWPRIVGEOGRAPHY */
	(iw_handler) rt2x00_ioctl_get_geography,		/* SIOCGIWPRIVGEOGRAPHY */
};

static const struct iw_priv_args rt2x00_ioctl_priv_args[] = {
	{ SIOCSIWPRIVNETWORK, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "set_network" },
	{ SIOCGIWPRIVNETWORK, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_network" },
	{ SIOCSIWPRIVPREAMBLE, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "set_preamble" },
	{ SIOCGIWPRIVPREAMBLE, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_preamble" },
	{ SIOCSIWPRIVANTENNA, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "set_antenna" },
	{ SIOCGIWPRIVANTENNA, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_antenna" },
	{ SIOCSIWPRIVGEOGRAPHY, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, 0, "set_geography" },
	{ SIOCGIWPRIVGEOGRAPHY, 0, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_geography" },
};

static const struct iw_handler_def rt2x00_ioctl_handler_def = {
	.num_standard		= ARRAY_SIZE(rt2x00_ioctl_handler),
	.num_private		= ARRAY_SIZE(rt2x00_ioctl_priv_handler),
	.num_private_args	= ARRAY_SIZE(rt2x00_ioctl_priv_args),
	.standard		= rt2x00_ioctl_handler,
	.private		= rt2x00_ioctl_priv_handler,
	.private_args		= rt2x00_ioctl_priv_args,
	.get_wireless_stats	= rt2x00_get_wireless_stats,
};

/*
 * TX/RX related routines.
 */
static int
rt2x00_start_xmit(struct ieee80211_txb *txb, struct net_device *net_dev, int pri)
{
	struct _rt2x00_device		*device = rt2x00_device(net_dev);
	struct _rt2x00_core		*core = rt2x00_core(device);
	struct sk_buff			*skb = NULL;
	u16				frame = 0x0000;
	u8				counter = 0x00;

	if(!test_bit(DEVICE_RADIO_ON, &device->flags))
		return -ENOTCONN;

	frame = rt2x00_get_sequence(device);

	spin_lock(&core->tx.lock);

	for(counter= 0x00; counter < txb->nr_frags; counter++){
		skb = txb->fragments[counter];
		txb->fragments[counter] = NULL;

		if(unlikely(!skb))
			break;

		skb->priority = pri;
		((struct ieee80211_hdr_3addr*)skb->data)->seq_ctl = cpu_to_le16(frame);
		((struct _skb_cb*)&skb->cb)->id = frame;
		((struct _skb_cb*)&skb->cb)->packets = txb->nr_frags;

		skb_queue_tail(&core->tx.queue, skb);

		frame += 0x0001;
	}

	if(core->tx.ready)
		queue_work(core->workqueue, &core->tx.work);

	spin_unlock(&core->tx.lock);

	ieee80211_txb_free(txb);

	return 0;
}

void
rt2x00_rx(struct _rt2x00_device *device, struct sk_buff *skb)
{
	struct _rt2x00_core	*core = rt2x00_core(device);

	skb_queue_tail(&core->rx.queue, skb);

	queue_work(core->workqueue, &core->rx.work);
}
EXPORT_SYMBOL_GPL(rt2x00_rx);

void
rt2x00_tx(struct _rt2x00_device *device, u8 ring_type)
{
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(!test_bit(DEVICE_RADIO_ON, &device->flags))
		return;

	if(ring_type & RING_TX)
		rt2x00_kick_ring(&core->tx);

	if(ring_type & RING_PRIO)
		rt2x00_kick_ring(&core->prio);

	if(ring_type & RING_ATIM)
		rt2x00_kick_ring(&core->atim);

	if(ring_type & RING_BEACON)
		rt2x00_kick_ring(&core->beacon);
}
EXPORT_SYMBOL_GPL(rt2x00_tx);

void
rt2x00_update_stats(struct _rt2x00_device *device, u8 flags, u8 value)
{
	struct _rt2x00_core	*core = rt2x00_core(device);

	if(flags & STATS_RX_CRC || flags & STATS_RX_PHYSICAL || flags & STATS_RX_QUALITY)
		core->ieee80211->stats.rx_crc_errors += value;

	if(flags & STATS_RX_DROP)
		core->ieee80211->stats.rx_dropped += value;

	if(flags & STATS_TX_RETRY_COUNT)
		core->ieee80211->stats.collisions += value;

	if(flags & STATS_TX_RESULT){
		switch(value){
			case TX_FAIL_RETRY:
				core->wireless_stats.discard.retries++;
				core->ieee80211->stats.tx_errors++;
			break;
			case TX_FAIL_INVALID:
			case TX_FAIL_OTHER:
				core->wireless_stats.discard.misc++;
				core->ieee80211->stats.tx_errors++;
			break;
		}
	}
}
EXPORT_SYMBOL_GPL(rt2x00_update_stats);

/*
 * Workqueue handlers.
 */
static void
rt2x00_delayed_rx(void *data)
{
	struct _packet_ring		*ring = data;
	struct _rt2x00_core		*core = ring->core;
	struct ieee80211_hdr_4addr	*ieee80211hdr = NULL;
	struct ieee80211_rx_stats	rx_stats;
	struct sk_buff			*skb = NULL;
	u16				frame = 0x00;

	down_read(&core->device->rt2x00_sem);

	rx_stats.rate = core->config.active.bitrate * 5000; /* in 100 kbps */
	rx_stats.received_channel = core->config.active.channel;
	if(rt2x00_is_24ghz_channel(core->config.active.channel))
		rx_stats.freq = IEEE80211_24GHZ_BAND;
	else
		rx_stats.freq = IEEE80211_52GHZ_BAND;

	up_read(&core->device->rt2x00_sem);

	spin_lock(&ring->lock);

	while(!skb_queue_empty(&ring->queue)){
		skb = skb_dequeue(&ring->queue);
		if(!skb){
			ERROR("Failed to dequeue packet.\n");
			break;
		}

		rx_stats.rssi = ((struct _skb_cb*)&skb->cb)->rssi;
		rx_stats.noise = ((struct _skb_cb*)&skb->cb)->noise;
		rx_stats.len = skb->len;

		ieee80211hdr = (struct ieee80211_hdr_4addr*)skb->data;
		frame = le16_to_cpu(ieee80211hdr->frame_ctl);

		if(WLAN_FC_GET_TYPE(frame) == IEEE80211_FTYPE_MGMT
		&& core->config.active.iw_mode != IW_MODE_MONITOR){
			if(WLAN_FC_GET_STYPE(frame) == IEEE80211_STYPE_ASSOC_REQ
			|| WLAN_FC_GET_STYPE(frame) == IEEE80211_STYPE_REASSOC_REQ)
				rt2x00_mgmt_rcv_assoc(core->device, ieee80211hdr, &rx_stats);
			else
				ieee80211_rx_mgt(core->ieee80211, ieee80211hdr, &rx_stats);
			dev_kfree_skb_any(skb);
		}else if(!ieee80211_rx(core->ieee80211, skb, &rx_stats))
			dev_kfree_skb_any(skb);
	}

	spin_unlock(&ring->lock);
}

static int
rt2x00_xmit_packet(struct _rt2x00_core *core, struct _packet_ring *ring)
{
	struct sk_buff			*skb = NULL;
	struct ieee80211_hdr_3addr	*ieee80211hdr = NULL;
	u16				xmit_flags = 0x0000;
	u8				frame_gap = 0x00;
	u8				rate = 0x00;

	if(ring->type != RING_PRIO
	&& test_bit(DEVICE_SCANNING, &core->device->flags)
	&& (core->scan_flags & IW_SCAN_ALL_FREQ))
		return -EBUSY;

	if(!ring->ready)
		return - EBUSY;

	skb = skb_dequeue(&ring->queue);
	if(!skb){
		ERROR("Failed to dequeue packet.\n");
		return -ENODATA;
	}

	ieee80211hdr = (struct ieee80211_hdr_3addr*)skb->data;

	if(ring->type == RING_BEACON)
		ieee80211hdr->seq_ctl = cpu_to_le16(rt2x00_get_sequence(core->device));

	if(WLAN_GET_SEQ_FRAG(ieee80211hdr->seq_ctl) == 1){
		frame_gap = XMIT_IFS_BACKOFF;
		xmit_flags |= XMIT_NEW_SEQUENCE;
	}else
		frame_gap = XMIT_IFS_SIFS;

	if(ieee80211hdr->frame_ctl & IEEE80211_STYPE_RTS){
		xmit_flags |= XMIT_RTS;
		frame_gap = XMIT_IFS_SIFS;
	}

	xmit_flags |= frame_gap;

	if(is_valid_ether_addr(&ieee80211hdr->addr1[0])
	&& ieee80211hdr->frame_ctl & IEEE80211_STYPE_ACK)
		xmit_flags |= XMIT_ACK;

	if(ieee80211hdr->frame_ctl & IEEE80211_STYPE_PROBE_RESP)
		xmit_flags |= XMIT_TIMESTAMP;

	down_read(&core->device->rt2x00_sem);

	rate = core->config.active.bitrate;

	if(core->config.active.flags & CONFIG_ENABLED_SHORT_PREAMBLE)
		xmit_flags |= XMIT_SHORT_PREAMBLE;

	if(core->config.active.flags & CONFIG_ENABLED_LONG_RETRY)
		xmit_flags |= XMIT_LONG_RETRY;

	up_read(&core->device->rt2x00_sem);

	if(rt2x00_is_ofdm_bitrate(rate))
		xmit_flags |= XMIT_OFDM;

	if(ring->type != RING_TX)
		xmit_flags |= XMIT_START;

	if(core->device->handler->dev_xmit_packet(core->device, skb, ring->type, rate, xmit_flags)){
		ring->ready = 0;
		skb_queue_head(&ring->queue, skb);
		return -ENOMEM;
	}

	if(ring->type != RING_BEACON)
		dev_kfree_skb_any(skb);
	else
		skb_queue_head(&ring->queue, skb);

	return 0;
}

static void
rt2x00_delayed_tx(void *data)
{
	struct _packet_ring	*ring = data;
	struct _rt2x00_device	*device = ring->core->device;
	struct sk_buff		*skb = NULL;
	u16			packet_id = 0x0000;

	spin_lock(&ring->lock);

	if(ring->type == RING_TX)
		goto dequeue_tx;

	while(!skb_queue_empty(&ring->queue)){
		if(rt2x00_xmit_packet(ring->core, ring))
			goto exit_kick_tx;
		if(ring->type == RING_BEACON)
			goto exit;
	}
	goto exit_kick_tx;

dequeue_tx:
	while(!skb_queue_empty(&ring->queue)){
		skb = skb_peek(&ring->queue);
		if(!skb)
			goto exit_kick_tx;

		packet_id = ((struct _skb_cb*)&skb->cb)->id;

		if(device->handler->dev_test_tx(device, ((struct _skb_cb*)&skb->cb)->packets))
			goto exit_kick_tx;

		do{
			if(rt2x00_xmit_packet(ring->core, ring))
				goto exit_kick_tx;

			skb = skb_peek(&ring->queue);
			if(!skb)
				goto exit_kick_tx;
		}while(((struct _skb_cb*)&skb->cb)->id == packet_id);
	}

exit_kick_tx:
	if(device->handler->dev_xmit_packet(device, NULL, ring->type, 0, XMIT_START))
		ERROR("Failed to kick TX queue.\n");

exit:
	spin_unlock(&ring->lock);
}

/*
 * net_device handlers.
 */
static int
rt2x00_open(struct net_device *net_dev)
{
	struct _rt2x00_device	*device = rt2x00_device(net_dev);

	if(!try_module_get(device->handler->dev_module)){
		ERROR("Failure to acquire module.\n");
		return -ENODEV;
	}

	if(test_and_set_bit(DEVICE_ENABLED, &device->flags)){
		ERROR("Device already enabled.\n");
		module_put(device->handler->dev_module);
		return -EBUSY;
	}

	rt2x00_radio_on(device);

	if(!test_bit(DEVICE_IRQ_ENABLED, &device->flags)){
		module_put(device->handler->dev_module);
		return -EIO;
	}

	return 0;
}

static int
rt2x00_stop(struct net_device *net_dev)
{
	struct _rt2x00_device	*device = rt2x00_device(net_dev);

	if(!test_and_clear_bit(DEVICE_ENABLED, &device->flags)){
		ERROR("Device already disabled.\n");
		return -EBUSY;
	}

	rt2x00_radio_off(device);

	module_put(device->handler->dev_module);

	return 0;
}

static int
rt2x00_change_mtu(struct net_device *net_dev, int new_mtu)
{
	if((new_mtu < 256) || (new_mtu > 2048))
		return -EINVAL;

	net_dev->mtu = new_mtu;

	return 0;
}

static int
rt2x00_set_mac_address(struct net_device *net_dev, void *addr)
{
	struct _rt2x00_device	*device = rt2x00_device(net_dev);
	struct sockaddr		*mac = (struct sockaddr*) addr;
	int			status = -ENOTSUPP;

	down_write(&device->rt2x00_sem);

	if(netif_running(net_dev)){
		status = -EBUSY;
		goto exit;
	}

	if(net_dev->addr_len != ETH_ALEN
	|| !is_valid_ether_addr(&mac->sa_data[0])){
		ERROR("Invalid MAC address.\n");
		status = -EINVAL;
		goto exit;
	}

	memcpy(&net_dev->dev_addr, &mac->sa_data, ETH_ALEN);
	status = device->handler->dev_set_mac(device, &mac->sa_data[0]);

exit:
	up_write(&device->rt2x00_sem);

	return status;
}

static void
rt2x00_tx_timeout(struct net_device *net_dev)
{
	struct _rt2x00_device	*device = ieee80211_priv(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	WARNING("Tx timed out. Start kicking all packet rings.\n");

	if(unlikely(device->handler->dev_xmit_packet(device, NULL, RING_TX | RING_PRIO | RING_ATIM, 0, XMIT_START)))
		ERROR("Kicking xmit failed.\n");

	rt2x00_kick_ring(&core->tx);
	rt2x00_kick_ring(&core->prio);
	rt2x00_kick_ring(&core->atim);
}

static struct net_device_stats*
rt2x00_get_stats(struct net_device *net_dev)
{
	return &((struct ieee80211_device*)netdev_priv(net_dev))->stats;
}

/*
 * Initialization handlers.
 */
static void
rt2x00_init_config(struct _rt2x00_core *core)
{
	memset(&core->config.nickn, 0x00, sizeof(core->config.nickn));
	memset(&core->config.user.essid, 0x00, sizeof(core->config.user.essid));
	memset(&core->config.user.bssid, 0x00, sizeof(core->config.user.bssid));

	core->config.user.network = 0;
	core->config.user.iw_mode = IW_MODE_AUTO;
	core->config.user.geography = IOCTL_GEOGRAPHY_INVAL;
	core->config.user.channel = 0;
	core->config.user.bitrate = 0;
	core->config.user.flags |= CONFIG_ACCEPT_UNICAST | CONFIG_ACCEPT_MULTICAST;
	core->config.user.short_retry = 4;
	core->config.user.long_retry = 7;
	core->config.user.rts_threshold = MAX_FRAG_THRESHOLD;
	core->config.user.txpower = 100;
	core->config.user.antenna_flags |= ANTENNA_TX_DIV | ANTENNA_RX_DIV;
#if WIRELESS_EXT > 17
	core->config.user.wpa_version = IW_AUTH_WPA_VERSION_DISABLED;
	core->config.user.mgmt_auth = 0;
#endif /* WIRELESS_EXT > 17 */
	core->config.beacon = 100;
	core->config.plcp = 48;
	core->config.sifs = 10;
	core->config.sequence = 0;
	core->config.user.sensitivity = 28;
	core->config.user_commit = COMMIT_ALL_CONFIG;
}

/*
 * Module handlers.
 */
struct net_device*
rt2x00_core_probe(struct _rt2x00_dev_handler *handler, void *priv, u32 sizeof_dev)
{
	struct net_device	*net_dev = NULL;
	struct _rt2x00_device	*device = NULL;
	struct _rt2x00_core	*core = NULL;

	if(unlikely(handler->dev_probe == NULL)
	|| unlikely(handler->dev_remove == NULL)
	|| unlikely(handler->dev_radio_on == NULL)
	|| unlikely(handler->dev_radio_off == NULL)
	|| unlikely(handler->dev_update_config == NULL)
	|| unlikely(handler->dev_read_mac == NULL)
	|| unlikely(handler->dev_set_mac == NULL)
	|| unlikely(handler->dev_test_tx == NULL)
	|| unlikely(handler->dev_xmit_packet == NULL)){
		ERROR("Not enough handlers provided by device.\n");
		goto exit;
	}

	net_dev = alloc_ieee80211(sizeof_dev + sizeof(*device) + sizeof(*core));
	if(!net_dev){
		ERROR("Not enough memory to allocate net_device.\n");
		goto exit;
	}

	device = ieee80211_priv(net_dev);
	memset(device, 0x00, sizeof(*device));

	core = (void*)device + sizeof(*device);
	core->ieee80211 = netdev_priv(net_dev);
	core->device = device;
	core->network = NULL;

	device->priv = core + sizeof(*core);
	device->owner = core;
	device->handler = handler;
	device->net_dev = net_dev;

	init_rwsem(&device->rt2x00_sem);

	core->workqueue = create_singlethread_workqueue("rt2x00");
	if (!core->workqueue)
		goto exit_free_ieee80211;
	INIT_WORK(&core->scan_tuner, rt2x00_scan_periodic, device);
	INIT_WORK(&core->link_tuner, rt2x00_link_periodic, device);

	rt2x00_init_ring(device, &core->rx, RING_RX, rt2x00_delayed_rx);
	rt2x00_init_ring(device, &core->tx, RING_TX, rt2x00_delayed_tx);
	rt2x00_init_ring(device, &core->atim, RING_ATIM, rt2x00_delayed_tx);
	rt2x00_init_ring(device, &core->prio, RING_PRIO, rt2x00_delayed_tx);
	rt2x00_init_ring(device, &core->beacon, RING_BEACON, rt2x00_delayed_tx);

	rt2x00_init_config(core);

	if(device->handler->dev_probe(device, &core->config, priv)){
		ERROR("Device probe failed.\n");
		goto exit_free_workqueue;
	}

	device->handler->dev_read_mac(device, net_dev);

	core->ieee80211->hard_start_xmit = rt2x00_start_xmit;
	core->ieee80211->handle_auth = rt2x00_handle_auth;
	core->ieee80211->handle_disassoc = rt2x00_handle_disassoc;
	core->ieee80211->handle_assoc_response = rt2x00_handle_assoc_response;

	core->ieee80211->config |= CFG_IEEE80211_RTS;
	core->ieee80211->state = IEEE80211_INITIALIZED;

	net_dev->open = rt2x00_open;
	net_dev->stop = rt2x00_stop;
	net_dev->change_mtu = rt2x00_change_mtu;
	net_dev->set_mac_address = rt2x00_set_mac_address;
	net_dev->tx_timeout = rt2x00_tx_timeout;
	net_dev->watchdog_timeo = 2 * HZ;
	net_dev->wireless_handlers = &rt2x00_ioctl_handler_def;
	net_dev->wireless_data = &core->wireless_data;

	net_dev->get_stats = rt2x00_get_stats;
	net_dev->get_wireless_stats = rt2x00_get_wireless_stats;

	if(register_netdev(net_dev)){
		ERROR("net_device registration failed.\n");
		goto exit_dev_remove;
	}

	set_bit(DEVICE_AWAKE, &device->flags);

	/*
	 * Activate current configuration.
	 */
	rt2x00_activate_config(device);

	return net_dev;

exit_dev_remove:
	device->handler->dev_remove(device);

exit_free_workqueue:
	destroy_workqueue(core->workqueue);

exit_free_ieee80211:
	free_ieee80211(net_dev);

exit:
	return NULL;
}
EXPORT_SYMBOL_GPL(rt2x00_core_probe);

void
rt2x00_core_remove(struct net_device *net_dev)
{
	struct _rt2x00_device	*device = rt2x00_device(net_dev);
	struct _rt2x00_core	*core = rt2x00_core(device);

	destroy_workqueue(core->workqueue);

	unregister_netdev(net_dev);

	down_write(&device->rt2x00_sem);

	device->handler->dev_remove(device);

	free_ieee80211(net_dev);

	up_write(&device->rt2x00_sem);
}
EXPORT_SYMBOL_GPL(rt2x00_core_remove);

/*
 * RT2x00 core module information.
 */
static char version[] = DRV_NAME " - " DRV_VERSION " (" DRV_RELDATE ") by " DRV_PROJECT;

int rt2x00_debug_level = 0x00000000;
EXPORT_SYMBOL_GPL(rt2x00_debug_level);

MODULE_AUTHOR(DRV_PROJECT);
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION("Ralink RT2400/RT2500 PCI, PCMCIA & USB Wireless LAN driver.");
MODULE_SUPPORTED_DEVICE("Ralink RT2460, RT2560 & RT2570 PCI, PCMCIA & USB chipset based cards");
MODULE_LICENSE("GPL");

static int __init rt2x00_core_init(void)
{
	printk(KERN_INFO "Loading module: %s\n", version);
	return 0;
}

static void __exit rt2x00_core_exit(void)
{
	printk(KERN_INFO "Unloading module: %s\n", version);
}

module_init(rt2x00_core_init);
module_exit(rt2x00_core_exit);
