/*
	Copyright (C) 2004 - 2005 rt2x00 SourceForge Project
	<http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
	Module Name: rt2x00dev.h

	Abstract: Data structures and registers for the rt2x00 device modules.
	Supported chipsets: RT2460, RT2560 & RT2570.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		3 Sep 2005	Initial Release (Beta 1).
 */

#ifndef RT2X00DEV_H
#define RT2X00DEV_H

#include <linux/dma-mapping.h>

/*
 * Register handlers.
 * The registers on the RT2x00 devices are little endian.
 * At all times should variables that must be written to the registers,
 * should be treated as little endian.
 * For that purpose we use the FIELD** defines, since they will use the 
 * appropriate _rt2x00_field** structures to store the values.
 * The 2 structures are being used by the rt2x00_set_field** and rt2x00_get_field**
 * functions to read/write the field at the correct position in the register.
 */
struct _rt2x00_field16{
	u16	bit_offset;
	u16	bit_mask;
} __attribute__ ((packed));

struct _rt2x00_field32{
	u32	bit_offset;
	u32	bit_mask;
} __attribute__ ((packed));

#define FIELD16(__offset, __mask)	( (struct _rt2x00_field16) { (__offset), (__mask) } )
#define FIELD32(__offset, __mask)	( (struct _rt2x00_field32) { (__offset), (__mask) } )

static inline void
rt2x00_set_field32(u32 *reg, const struct _rt2x00_field32 field, const u32 value)
{
	*reg &= cpu_to_le32(~(field.bit_mask));
	*reg |= cpu_to_le32((value << field.bit_offset) & field.bit_mask);
}

static inline u32
rt2x00_get_field32(const u32 reg, const struct _rt2x00_field32 field)
{
	return le32_to_cpu((reg & field.bit_mask) >> field.bit_offset);
}

static inline void
rt2x00_set_field16(u16 *reg, const struct _rt2x00_field16 field, const u16 value)
{
	*reg &= cpu_to_le16(~(field.bit_mask));
	*reg |= cpu_to_le16((value << field.bit_offset) & field.bit_mask);
}

static inline u16
rt2x00_get_field16(const u16 reg, const struct _rt2x00_field16 field)
{
	return le16_to_cpu((reg & field.bit_mask) >> field.bit_offset);
}

/*
 * rf register sructure for channel selection.
 */
struct _rf_channel{
	u32				rf1;
	u32				rf2;
	u32				rf3;
	u32				rf4;
}__attribute__ ((packed));

/*
 * Chipset identification
 * The chipset on the device is composed of a RT and RF chip.
 * The chipset combination is important for determining device capabilities.
 */
struct _rt2x00_chip{
	u16				rt;
	u16				rf;
} __attribute__ ((packed));

/*
 * Set chipset data.
 * Some rf values for RT2400 devices are equal to rf values for RT2500 devices.
 * To prevent problems, all rf values will be masked to clearly seperate each chipset.
 */
static inline void
set_chip(struct _rt2x00_chip *chipset, const u16 rt, const u16 rf)
{
	INFO("Chipset detected - rt: %04x, rf: %04x.\n", rt, rf);

	chipset->rt = rt;
	chipset->rf = rf | (chipset->rt & 0xff00);
}

static inline char
rt2x00_rt(const struct _rt2x00_chip *chipset, const u16 chip)
{
	return (chipset->rt == chip);
}

static inline char
rt2x00_rf(const struct _rt2x00_chip *chipset, const u16 chip)
{
	return (chipset->rf == chip);
}

/*
 * _data_ring
 * Data rings are used by the device to send and receive packets.
 * The data_addr is the base address of the data memory.
 * Device specifice information is pointed to by the priv pointer.
 * The index values may only be changed with the functions ring_index_inc()
 * and ring_index_done_inc().
 */
struct _data_ring{
	/*
	 * Base address of packet ring.
	 */
	dma_addr_t			data_dma;
	void				*data_addr;

	/*
	 * Private device specific data.
	 */
	void				*priv;
	struct _rt2x00_device		*device;

	/*
	 * Current index values.
	 */
	u8				index;
	u8				index_done;

	/*
	 * Ring type set with RING_* define.
	 */
	u8				ring_type;

	/*
	 * Number of entries in this ring.
	 */
	u8				max_entries;

	/*
	 * Size of packet and descriptor in bytes.
	 */
	u16				entry_size;
	u16				desc_size;

	/*
	 * Total allocated memory size.
	 */
	u32				mem_size;
} __attribute__ ((packed));

/*
 * Number of entries in a packet ring.
 */
#define RX_ENTRIES			16
#define TX_ENTRIES			16
#define ATIM_ENTRIES			1
#define PRIO_ENTRIES			2
#define BEACON_ENTRIES			1

/*
 * Size of a single entry in the packet ring for packet data.
 */
#define RX_ENTRY_SIZE			2048
#define TX_ENTRY_SIZE			2048
#define ATIM_ENTRY_SIZE			256
#define PRIO_ENTRY_SIZE			256
#define BEACON_ENTRY_SIZE		256

/*
 * Initialization and cleanup routines.
 */
static inline int
rt2x00_alloc_ring(
	struct _rt2x00_device *device,
	struct _data_ring *ring,
	struct device *dev,
	const u8 ring_type,
	const u16 max_entries,
	const u16 entry_size,
	const u16 desc_size)
{
	ring->device = device;
	ring->index = 0;
	ring->index_done = 0;
	ring->ring_type = ring_type;
	ring->max_entries = max_entries;
	ring->entry_size = entry_size;
	ring->desc_size = desc_size;
	ring->mem_size = ring->max_entries * (ring->desc_size + ring->entry_size);

	ring->data_addr = dma_alloc_coherent(dev, ring->mem_size, &ring->data_dma, GFP_KERNEL);
	if(!ring->data_addr){
		ring->mem_size = 0;
		return -ENOMEM;
	}

	memset(ring->data_addr, 0x00, ring->mem_size);

	return 0;
}

static inline void
rt2x00_free_ring(struct _data_ring *ring, struct device *dev)
{
	if(ring->data_addr && ring->mem_size)
		dma_free_coherent(dev, ring->mem_size, ring->data_addr, ring->data_dma);

	ring->data_addr = NULL;
	ring->index = 0;
	ring->index_done = 0;
	ring->max_entries = 0;
	ring->mem_size = 0;
}

/*
 * Ring index manipulation functions.
 */
static inline void
rt2x00_ring_index_inc(struct _data_ring *ring)
{
	ring->index = (++ring->index < ring->max_entries) ? ring->index : 0;
}

static inline void
rt2x00_ring_index_done_inc(struct _data_ring *ring)
{
	ring->index_done = (++ring->index_done < ring->max_entries) ? ring->index_done : 0;
}

static inline u8
rt2x00_ring_free_entries(struct _data_ring *ring)
{
	if(ring->index > ring->index_done)
		return ring->max_entries - (ring->index - ring->index_done);
	else
		return ring->index_done - ring->index;
}

static inline int
rt2x00_ring_rx_packet(struct _rt2x00_device *device, u16 size, void *data, u16 rssi, u16 noise)
{
	struct sk_buff		*skb = NULL;

	skb = dev_alloc_skb(size + NET_IP_ALIGN);
	if(!skb){
		rt2x00_update_stats(device, STATS_RX_DROP, 1);
		return -ENOMEM;
	}

	skb_reserve(skb, NET_IP_ALIGN);

	memcpy(skb_put(skb, size), data, size);

	((struct _skb_cb *)(skb->cb))->rssi = rssi;
	((struct _skb_cb *)(skb->cb))->noise = noise;

	rt2x00_rx(device, skb);

	return 0;
}

/*
 * Return PLCP value matching the rate.
 * PLCP values according to ieee802.11a-1999 p.14.
 */
static inline u8
rt2x00_get_plcp(const u8 rate)
{
	u8	counter = 0x00;
	u8	plcp[12] = {
		0x00, 0x01, 0x02, 0x03,					/* CCK. */
		0x0b, 0x0f, 0x0a, 0x0e, 0x09, 0x0d, 0x08, 0x0c,		/* OFDM. */
	};

	for(; counter < 12; counter++){
		if(capabilities.bitrate[counter] == rate)
			return plcp[counter];
	}

	return 0xff;
}

#define GET_PREAMBLE(__config)	( ((__config)->active.flags & CONFIG_ENABLED_SHORT_PREAMBLE) ? 72 :  144 )

#endif /* RT2X00DEV_H */
