/*
	Copyright (C) 2004 - 2006 rt2x00 SourceForge Project
	<http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
	Module: rt2x00debug
	Abstract: rt2x00debug specific routines.
	Supported chipsets: RT2460, RT2560, RT2570,
	rt2561, rt2561s, rt2661 & rt2573.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/debugfs.h>

#include <asm/div64.h>
#include <asm/uaccess.h>

#include "rt2x00debug.h"

/*
 * For displaying the register values, all bytes will be formatted.
 * To do this additional bytes will be attached to the byte string.
 * The intended displaying format for register values is:
 *
 * 0001 0xffffffff
 * 0002 0xffffffff
 *
 * The intended displaying format for eeprom values is:
 *
 * 0001 0xffff
 * 0002 0xffff
 *
 * The first value is the word index number, and the second is the
 * hexadecimal output of the value. The length of that strings depends
 * on the length of the register or eeprom.
 */
#define PRINT16_STR		( "%.4d 0x%.4x\n" )
#define PRINT32_STR		( "%.4d 0x%.8x\n" )
#define PRINT_LENGTH(__wsize)	( 8 + (2 * (__wsize)) )
#define PRINT_MAX_LENGTH	( PRINT_LENGTH(sizeof(u32)) )

/*
 * If this field is enabled, debugfs has been disabled in the kernel,
 * and we should not return errors to the user about the failed registration.
 */
static unsigned int rt2x00_debugfs_disabled = 0;

/*
 * Main rt2x00 entry, this will create the "rt2x00" folder at the root
 * of the debugfs folder.
 */
static struct dentry *rt2x00_entry;

struct rt2x00debug_intf {
	/*
	 * Reference to the rt2x00debug structure
	 * which can be used to communicate with
	 * the registers.
	 */
	struct rt2x00debug *debug;

	/*
	 * debugfs entry for the interface.
	 */
	struct dentry *intf_entry;

	/*
	 * debugfs entry for the register.
	 */
	struct dentry *reg_entry;

	/*
	 * debugfs entry for the eeprom.
	 */
	struct dentry *eeprom_entry;
};

static ssize_t rt2x00debug_file_read(void *device,
	char __user *buf, size_t length, loff_t *offset, debug_read_t *read,
	unsigned int reg_length, unsigned int word_length)
{
	loff_t tmp = *offset;
	unsigned int print_size;
	unsigned int total_words;
	unsigned int read_words;
	unsigned int words;
	unsigned int i;
	void *data;
	char line[PRINT_MAX_LENGTH];

	/*
	 * Make the required calculations for
	 * - The length (in bytes) of a single line in the file.
	 * - Total number of words inside the register.
	 * - Total number of words that have been read.
	 */
	print_size = PRINT_LENGTH(word_length);
	total_words = reg_length / word_length;
	do_div(tmp, print_size);
	read_words = tmp;

	/*
	 * Make sure we only read the data that is available to us,
	 * which basically means we shouldn't read beyond the maximum
	 * number of words available.
	 */
	if (read_words >= total_words)
		return 0;

	/*
	 * Besides the register information we want to append
	 * additional bytes used to format the data to a readable format.
	 * This means we need to calculate how much words we should read,
	 * and how much extra bytes can be appended without exceeding the
	 * number of bytes that the kernel offered us to fill.
	 */
	words = total_words - read_words;
	if ((words * print_size) > length)
		words = length / print_size;
	length = words * word_length;

	/*
	 * Create temporary buffer.
	 */
	data = kzalloc(length, GFP_KERNEL);
	if (!data)
		return -ENOMEM;

	/*
	 * Read the requested information from the device.
	 */
	read(device, *offset, data, length);

	/*
	 * To make sure we will get the correct readable output,
	 * we have to convert the values into a string.
	 */
	for (i = 0; i < words; i++) {
		if (word_length == sizeof(u16))
			sprintf(&line[0], PRINT16_STR, read_words,
				((u16*)data)[i]);
		else
			sprintf(&line[0], PRINT32_STR, read_words,
				((u32*)data)[i]);

		if (copy_to_user(buf, &line, print_size))
			goto exit;

		read_words++;
		buf += print_size;
	}

	/*
	 * Free the temporary buffer.
	 */
	kfree(data);

	/*
	 * Update the offset, and return the length that has been read.
	 */
	*offset += (words * print_size);
	return (words * print_size);

exit:
	kfree(data);

	return -EFAULT;
}

static int rt2x00debug_file_open(struct inode *inode, struct file *file)
{
	struct rt2x00debug_intf *intf;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
	file->private_data = inode->i_private;
#else
	file->private_data = inode->u.generic_ip;
#endif

	intf = (struct rt2x00debug_intf*)file->private_data;

	if (!try_module_get(intf->debug->owner))
		return -EBUSY;

	return 0;
}

static int rt2x00debug_file_release(struct inode *inode, struct file *file)
{
	struct rt2x00debug_intf *intf =
		(struct rt2x00debug_intf*)file->private_data;

	module_put(intf->debug->owner);

	return 0;
}

static ssize_t rt2x00debug_reg_read(struct file *file,
	char __user *buf, size_t length, loff_t *offset)
{
	struct rt2x00debug_intf *intf =
		(struct rt2x00debug_intf*)file->private_data;
	struct rt2x00debug *debug = intf->debug;

	return rt2x00debug_file_read(debug->dev,
		buf, length, offset, debug->read_reg,
		debug->reg_length, debug->reg_word_length);
}

static const struct file_operations reg_fops = {
	.owner		= THIS_MODULE,
	.read		= rt2x00debug_reg_read,
	.open		= rt2x00debug_file_open,
	.release	= rt2x00debug_file_release,
};

static ssize_t rt2x00debug_eeprom_read(struct file *file,
	char __user *buf, size_t length, loff_t *offset)
{
	struct rt2x00debug_intf *intf =
		(struct rt2x00debug_intf*)file->private_data;
	struct rt2x00debug *debug = intf->debug;

	return rt2x00debug_file_read(debug->dev,
		buf, length, offset, debug->read_eeprom,
		debug->eeprom_length, sizeof(u16));
}

static const struct file_operations eeprom_fops = {
	.owner		= THIS_MODULE,
	.read		= rt2x00debug_eeprom_read,
	.open		= rt2x00debug_file_open,
	.release	= rt2x00debug_file_release,
};

int rt2x00debug_register(struct rt2x00debug *debug)
{
	struct rt2x00debug_intf *intf;

	if (rt2x00_debugfs_disabled)
		return 0;

	intf = kzalloc(sizeof(struct rt2x00debug_intf), GFP_KERNEL);
	if (!intf)
		return -ENOMEM;

	intf->intf_entry = debugfs_create_dir(debug->intf_name, rt2x00_entry);
	if (IS_ERR(intf->intf_entry))
		goto exit;

	intf->reg_entry = debugfs_create_file("register", S_IRUGO,
		intf->intf_entry, intf, &reg_fops);
	if (IS_ERR(intf->reg_entry))
		goto exit_dir;

	intf->eeprom_entry = debugfs_create_file("eeprom", S_IRUGO,
		intf->intf_entry, intf, &eeprom_fops);
	if (IS_ERR(intf->eeprom_entry))
		goto exit_reg;

	intf->debug = debug;
	debug->priv = intf;

	return 0;

exit_reg:
	debugfs_remove(intf->reg_entry);

exit_dir:
	debugfs_remove(intf->intf_entry);

exit:
	kfree(intf);

	return -ENOMEM;
}
EXPORT_SYMBOL_GPL(rt2x00debug_register);

void rt2x00debug_deregister(struct rt2x00debug *debug)
{
	struct rt2x00debug_intf *intf = (struct rt2x00debug_intf*)debug->priv;

	if (rt2x00_debugfs_disabled || !intf)
		return;

	debugfs_remove(intf->eeprom_entry);
	debugfs_remove(intf->reg_entry);
	debugfs_remove(intf->intf_entry);
	kfree(intf);
}
EXPORT_SYMBOL_GPL(rt2x00debug_deregister);

/*
 * rt2x00debug module information.
 */
MODULE_AUTHOR("http://rt2x00.serialmonkey.com");
MODULE_VERSION("CVS");
MODULE_DESCRIPTION("Debugfs support for rt2x00 drivers.");
MODULE_LICENSE("GPL");

static int __init rt2x00debug_init(void)
{
	rt2x00_entry = debugfs_create_dir("rt2x00", NULL);
	if (IS_ERR(rt2x00_entry)) {
		if (PTR_ERR(rt2x00_entry) == -ENODEV)
			rt2x00_debugfs_disabled = 1;
		else
			return PTR_ERR(rt2x00_entry);
	}

	return 0;
}

static void __exit rt2x00debug_exit(void)
{
	if (!rt2x00_debugfs_disabled)
		debugfs_remove(rt2x00_entry);
}

module_init(rt2x00debug_init);
module_exit(rt2x00debug_exit);
